import pytest
import numpy as np
from pysersic import FitSingle,FitMulti, priors, rendering
from astropy.convolution import Gaussian2DKernel
from jax.random import PRNGKey
import arviz
from numpyro import distributions as dist,infer,sample, optim
from pysersic.pysersic import train_numpyro_svi_early_stop
from pysersic.priors import estimate_sky




im_s = np.zeros((40,40))
rng = np.random.default_rng(seed=10)
im_s += rng.normal(scale = 0.05, size = (40,40))
rms = np.ones(im_s.shape)*0.05
psf = Gaussian2DKernel(x_stddev=2.5).array
renderer = rendering.HybridRenderer((40,40), psf)
im_s = im_s + renderer.render_source( dict(xc = 20.,yc=20., flux = 200.), 'pointsource' )


@pytest.mark.parametrize('sky_type', ['none', 'flat', 'tilted-plane'])
def test_FitSingle_map(sky_type):
        props = priors.SourceProperties(im_s)
        
        prior = props.generate_prior(profile_type='pointsource', sky_type=sky_type)
        print (prior)
        fitter_single = FitSingle(im_s,rms,psf,prior)

        map_dict = fitter_single.find_MAP(rkey = PRNGKey(10))
        assert map_dict['xc'] == pytest.approx(20.017, rel = 1e-3)        
        assert map_dict['yc'] == pytest.approx(19.996, rel = 1e-3)
        assert map_dict['flux'] == pytest.approx(199.418, rel = 5e-2)

@pytest.mark.parametrize('method',['laplace','svi-mvn','svi-flow'])
def test_FitSingle_posterior(method):
        props = priors.SourceProperties(im_s)
        prior = props.generate_prior(profile_type='pointsource')

        fitter_single = FitSingle(im_s,rms,psf,prior)

        res = fitter_single.estimate_posterior(method = method, rkey = PRNGKey(3))
        post_sum = res.summary()

        assert post_sum['mean']['flux'] == pytest.approx(199.4, rel = 1e-2)
        assert post_sum['sd']['flux'] == pytest.approx(0.44, rel = 1e-1)


        assert post_sum['mean']['xc'] == pytest.approx(20.02, rel = 1e-2)
        assert post_sum['sd']['xc'] == pytest.approx(0.0085, rel = 2e-1)
        
        assert post_sum['mean']['yc'] == pytest.approx(19.99, rel = 1e-2)
        assert post_sum['sd']['yc'] ==  pytest.approx(0.008, rel = 1e-1)


def test_FitSingle_sample():
        props = priors.SourceProperties(im_s)
        prior = props.generate_prior(profile_type='pointsource')

        fitter_single = FitSingle(im_s,rms,psf,prior)

        res = fitter_single.sample(num_samples = 500,num_warmup = 500, num_chains = 1,rkey = PRNGKey(5))
        post_sum = arviz.summary(res.idata)

        assert post_sum['mean']['flux'] == pytest.approx(199.4, rel = 5e-2)
        assert post_sum['sd']['flux'] == pytest.approx(0.44, rel = 5e-2)

        assert post_sum['mean']['xc'] == pytest.approx(20.02, rel = 1e-2)
        assert post_sum['sd']['xc'] == pytest.approx(0.007, rel = 2e-1)
        
        assert post_sum['mean']['yc'] == pytest.approx(19.99, rel = 1e-2)
        assert post_sum['sd']['yc'] ==  pytest.approx(0.0085, rel = 1e-1)


im_m = np.zeros((40,40))
rng = np.random.default_rng(seed=12)
im_m += rng.normal(scale = 0.05, size = (40,40))
rms = np.ones(im_m.shape)*0.05
psf = Gaussian2DKernel(x_stddev=2.5).array
renderer = rendering.HybridRenderer((40,40), psf)
im_m = im_m + renderer.render_source( dict(xc = 10.,yc = 30., flux  = 150.), 'pointsource' )
im_m = im_m + renderer.render_source( dict(yc = 10.,xc = 30., flux  = 150.), 'pointsource' )


cat = {}
cat['x'] = [10.,30.]
cat['y'] = [30.,10.]
cat['r']= [0,0]
cat['flux'] = [150.,150.]
cat['type'] = ['pointsource','pointsource']
mp = priors.PySersicMultiPrior(catalog = cat, sky_type='none')
multi_fitter = FitMulti(im_m,rms,psf, mp)

sky_guess, sky_guess_err, n_pix_sky = estimate_sky(im_m)
@pytest.mark.parametrize('sky_type', ['none', 'flat', 'tilted-plane'])
def test_FitMulti_map(sky_type):
        mp = priors.PySersicMultiPrior(catalog = cat, sky_type=sky_type,sky_guess  = sky_guess, sky_guess_err=sky_guess_err)
        multi_fitter = FitMulti(im_m,rms,psf, mp)

        map_dict = multi_fitter.find_MAP(rkey = PRNGKey(10))

        assert map_dict['source_0']['xc'] == pytest.approx(10., rel = 1e-3)        
        assert map_dict['source_0']['yc'] == pytest.approx(30., rel = 1e-3)
        assert map_dict['source_0']['flux'] == pytest.approx(150.4, rel = 1e-3)

        assert map_dict['source_1']['xc'] == pytest.approx(30., rel = 1e-3)        
        assert map_dict['source_1']['yc'] == pytest.approx(10., rel = 1e-3)
        assert map_dict['source_1']['flux'] == pytest.approx(150., rel = 1e-3)

@pytest.mark.parametrize('method',['laplace','svi-mvn','svi-flow'])
def test_FitMulti_posterior(method):
        res = multi_fitter.estimate_posterior(method = method, rkey = PRNGKey(3))
        post_sum = res.summary()
        assert post_sum['mean']['flux_0'] == pytest.approx(150.4, rel = 1e-2)
        assert post_sum['sd']['flux_0'] == pytest.approx(0.45, rel = 2e-1)
        assert post_sum['mean']['flux_1'] == pytest.approx(150., rel = 1e-2)
        assert post_sum['sd']['flux_1'] == pytest.approx(0.45, rel = 1e-1)

        assert post_sum['mean']['xc_0'] == pytest.approx(10.0, rel = 1e-2)
        assert post_sum['sd']['xc_0'] == pytest.approx(0.01, rel = 1e-1)
        assert post_sum['mean']['xc_1'] == pytest.approx(30., rel = 1e-2)
        assert post_sum['sd']['xc_1'] == pytest.approx(0.01, rel = 1e-1)

        assert post_sum['mean']['yc_0'] == pytest.approx(30.0, rel = 1e-2)
        assert post_sum['sd']['yc_0'] == pytest.approx(0.01, rel = 2e-1)
        assert post_sum['mean']['yc_1'] == pytest.approx(10., rel = 1e-2)
        assert post_sum['sd']['yc_1'] == pytest.approx(0.01, rel = 2e-1)

def test_FitMulti_sample():
        res = multi_fitter.sample(num_samples = 500,num_warmup = 500, num_chains = 1,rkey = PRNGKey(5))
        post_sum = arviz.summary(res.idata)

        assert post_sum['mean']['flux_0'] == pytest.approx(150.4, rel = 1e-2)
        assert post_sum['sd']['flux_0'] == pytest.approx(0.44, rel =1e-1)
        assert post_sum['mean']['flux_1'] == pytest.approx(150.1, rel = 1e-2)
        assert post_sum['sd']['flux_1'] == pytest.approx(0.42, rel = 1e-1)

        assert post_sum['mean']['xc_0'] == pytest.approx(10.0, rel = 1e-2)
        assert post_sum['sd']['xc_0'] == pytest.approx(0.01, rel = 2e-1)
        assert post_sum['mean']['xc_1'] == pytest.approx(30., rel = 1e-2)
        assert post_sum['sd']['xc_1'] == pytest.approx(0.01, rel = 1e-1)

        assert post_sum['mean']['yc_0'] == pytest.approx(30.0, rel = 1e-2)
        assert post_sum['sd']['yc_0'] == pytest.approx(0.01, rel = 1e-1)
        assert post_sum['mean']['yc_1'] == pytest.approx(10., rel = 1e-2)
        assert post_sum['sd']['yc_1'] == pytest.approx(0.01, rel = 1e-1)

def test_train_numpyro_svi():
    def model():
        a = sample('a', dist.Normal())

    guide =infer.autoguide.AutoDelta(model)
    svi_kernel = infer.SVI(model,guide, optim.Adam(0.001), loss = infer.Trace_ELBO())
    res = train_numpyro_svi_early_stop(svi_kernel)
    assert float(res.params['a_auto_loc']) == pytest.approx(0, abs = 1e-2)


import warnings
from abc import abstractmethod
from typing import Iterable, Optional, Tuple, Union

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from interpax import interp1d
from scipy.special import comb

from .exceptions import *

base_profile_types = [
    "sersic",
    "doublesersic",
    "sersic_pointsource",
    "pointsource",
    "exp",
    "dev",
]
base_profile_params = dict(
    zip(
        base_profile_types,
        [
            ["xc", "yc", "flux", "r_eff", "n", "ellip", "theta"],
            [
                "xc",
                "yc",
                "flux",
                "f_1",
                "r_eff_1",
                "n_1",
                "ellip_1",
                "r_eff_2",
                "n_2",
                "ellip_2",
                "theta",
            ],
            ["xc", "yc", "flux", "f_ps", "r_eff", "n", "ellip", "theta"],
            ["xc", "yc", "flux"],
            ["xc", "yc", "flux", "r_eff", "ellip", "theta"],
            ["xc", "yc", "flux", "r_eff", "ellip", "theta"],
        ],
    )
)


class BaseRenderer(eqx.Module):
    im_shape: tuple = eqx.field(static=True)
    psf_shape: tuple = eqx.field(static=True)
    fft_shape: tuple = eqx.field(static=True)
    profile_func_dict: dict = eqx.field(static=True)

    pixel_PSF: jax.numpy.array
    PSF_fft: jax.numpy.array
    X: jax.numpy.array
    Y: jax.numpy.array
    FX: jax.numpy.array
    FY: jax.numpy.array
    x_mid: float
    y_mid: float
    fft_zeros: jnp.array
    img_zeros: jnp.array

    def __init__(self, im_shape: Iterable, pixel_PSF: jax.numpy.array) -> None:
        """Base class for different Renderers

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        """
        self.im_shape = im_shape
        self.pixel_PSF = pixel_PSF
        if not jnp.isclose(jnp.sum(self.pixel_PSF), 1.0, 0.1):
            warnings.warn(
                "PSF does not appear to be appropriately normalized; Sum(psf) is more than 0.1 away from 1."
            )
        self.psf_shape = jnp.shape(self.pixel_PSF)
        if jnp.any(self.im_shape < self.psf_shape):
            raise KernelError(
                "PSF pixel image size must be smaller than science image."
            )
        x = jnp.arange(self.im_shape[0])
        y = jnp.arange(self.im_shape[1])
        self.X, self.Y = jnp.meshgrid(x, y)
        self.x_mid = self.im_shape[0] / 2.0 - 0.5
        self.y_mid = self.im_shape[1] / 2.0 - 0.5

        # Set up pre-FFTed PSF
        f1d1 = jnp.fft.rfftfreq(self.im_shape[0])
        f1d2 = jnp.fft.fftfreq(self.im_shape[1])
        self.FX, self.FY = jnp.meshgrid(f1d1, f1d2)
        self.fft_shape = self.FX.shape
        fft_shift_arr_x = jnp.exp(
            jax.lax.complex(0.0, -1.0)
            * 2.0
            * 3.1415
            * -1
            * (self.psf_shape[0] / 2.0 - 0.5)
            * self.FX
        )
        fft_shift_arr_y = jnp.exp(
            jax.lax.complex(0.0, -1.0)
            * 2.0
            * 3.1415
            * -1
            * (self.psf_shape[1] / 2.0 - 0.5)
            * self.FY
        )
        self.PSF_fft = (
            jnp.fft.rfft2(self.pixel_PSF, s=self.im_shape)
            * fft_shift_arr_x
            * fft_shift_arr_y
        )

        # All the  renderers here these profile types
        self.profile_func_dict = {}
        for profile_type in base_profile_types:
            self.profile_func_dict[profile_type] = getattr(
                self, f"render_{profile_type}"
            )

        self.fft_zeros = jnp.zeros(self.fft_shape)
        self.img_zeros = jnp.zeros(self.im_shape)

    def conv_img(self, image):
        img_fft = jnp.fft.rfft2(image)
        conv_fft = img_fft * self.PSF_fft
        conv_im = jnp.fft.irfft2(conv_fft, s=self.im_shape)
        return conv_im

    def conv_fft(self, F_im):
        im = jnp.fft.irfft2(F_im * self.PSF_fft, s=self.im_shape)
        return im

    def combine_scene(self, F_im, int_im, obs_im):
        return self.conv_fft(F_im) + self.conv_img(int_im) + obs_im

    @abstractmethod
    def render_sersic(self, params: dict):
        return NotImplementedError

    def render_doublesersic(self, params: dict):
        dict_1 = {
            "xc": params["xc"],
            "yc": params["yc"],
            "flux": params["flux"] * params["f_1"],
            "n": params["n_1"],
            "ellip": params["ellip_1"],
            "theta": params["theta"],
            "r_eff": params["r_eff_1"],
        }
        dict_2 = {
            "xc": params["xc"],
            "yc": params["yc"],
            "flux": params["flux"] * (1.0 - params["f_1"]),
            "n": params["n_2"],
            "ellip": params["ellip_2"],
            "theta": params["theta"],
            "r_eff": params["r_eff_2"],
        }
        F1, im_int_1, im_obs_1 = self.render_sersic(dict_1)
        F2, im_int_2, im_obs_2 = self.render_sersic(dict_2)

        return F1 + F2, im_int_1 + im_int_2, im_obs_1 + im_obs_2

    def render_sersic_pointsource(self, params: dict):
        pointsource_dict = {}
        sersic_dict = params.copy()

        sersic_dict["flux"] = (1.0 - params["f_ps"]) * params["flux"]
        pointsource_dict["flux"] = sersic_dict.pop("f_ps") * params["flux"]
        pointsource_dict["xc"] = sersic_dict["xc"]
        pointsource_dict["yc"] = sersic_dict["yc"]

        F1, im_int_1, im_obs_1 = self.render_sersic(sersic_dict)
        F2, im_int_2, im_obs_2 = self.render_pointsource(pointsource_dict)
        return F1 + F2, im_int_1 + im_int_2, im_obs_1 + im_obs_2

    @abstractmethod
    def render_pointsource(self, params: dict):
        return NotImplementedError

    def render_exp(self, params: dict) -> jax.numpy.array:
        """Thin wrapper for an exponential profile based on render_sersic

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Exponential model
        """
        to_sersic = dict(params, n=1.0)
        return self.render_sersic(to_sersic)

    def render_dev(self, params: dict) -> jax.numpy.array:
        """Thin wrapper for a De Vaucouleurs profile based on render_sersic

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Exponential model
        """
        to_sersic = dict(params, n=4.0)
        return self.render_sersic(to_sersic)

    def render_for_model(self, param_dict, types, suffix):
        F_tot = jnp.zeros(self.fft_shape)
        int_im_tot = jnp.zeros(self.im_shape)
        obs_im_tot = jnp.zeros(self.im_shape)

        for j, prof_type in enumerate(types):
            new_dict = {
                param: param_dict[param + f"_{j:d}{suffix}"]
                for param in base_profile_params[prof_type]
            }
            F_cur, int_im_cur, obs_im_cur = self.profile_func_dict[prof_type](new_dict)
            F_tot = F_tot + F_cur
            int_im_tot = int_im_tot + int_im_cur
            obs_im_tot = obs_im_tot + obs_im_cur

        return self.combine_scene(F_tot, int_im_tot, obs_im_tot)

    def render_source(
        self, params: dict, profile_type: str, suffix: Optional[str] = ""
    ) -> jax.numpy.array:
        """Render an observed source of a given type and parameters

        Parameters
        ----------
        params : jax.numpy.array
            Parameters specifying the source
        profile_type : str
            Type of profile to use

        Returns
        -------
        jax.numpy.array
            Rendered, observed model
        """
        to_func = {k.replace(suffix, ""): v for k, v in params.items()}
        model_im = self.combine_scene(*self.profile_func_dict[profile_type](to_func))
        return model_im


class PixelRenderer(BaseRenderer):
    """
    Render class based on rendering in pixel space and then convolving with the PSF
    """

    os_pixel_size: int = eqx.field(static=True)
    num_os: int = eqx.field(static=True)

    w_os: jnp.array
    x_os_lo: int
    x_os_hi: int
    y_os_lo: int
    y_os_hi: int

    X_os: jnp.array
    Y_os: jnp.array

    def __init__(
        self,
        im_shape: Iterable,
        pixel_PSF: jax.numpy.array,
        os_pixel_size: Optional[int] = 6,
        num_os: Optional[int] = 12,
    ) -> None:
        """Initialize the PixelRenderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        os_pixel_size : Optional[int], optional
            Size of box around the center of the image to perform pixel oversampling
        num_os : Optional[int], optional
            Number of points to oversample by in each direction, by default 8
        """
        super().__init__(im_shape, pixel_PSF)
        self.os_pixel_size = os_pixel_size
        self.num_os = num_os

        # Use Gauss-Legendre coefficents for better integration when oversampling
        dx, w = jnp.array(np.polynomial.legendre.leggauss(self.num_os))
        w = w / 2.0
        dx = dx / 2.0

        # dx = np.linspace(-0.5,0.5, num= num_os,endpoint=True)
        # w = np.ones_like(dx)
        dx_os, dy_os = jnp.meshgrid(dx, dx)

        w1, w2 = jnp.meshgrid(w, w)
        self.w_os = w1 * w2

        i_mid = int(self.im_shape[0] / 2)
        j_mid = int(self.im_shape[1] / 2)

        self.x_os_lo, self.x_os_hi = (
            i_mid - self.os_pixel_size,
            i_mid + self.os_pixel_size,
        )
        self.y_os_lo, self.y_os_hi = (
            j_mid - self.os_pixel_size,
            j_mid + self.os_pixel_size,
        )

        self.X_os = (
            self.X[
                self.x_os_lo : self.x_os_hi,
                self.y_os_lo : self.y_os_hi,
                jnp.newaxis,
                jnp.newaxis,
            ]
            + dx_os
        )
        self.Y_os = (
            self.Y[
                self.x_os_lo : self.x_os_hi,
                self.y_os_lo : self.y_os_hi,
                jnp.newaxis,
                jnp.newaxis,
            ]
            + dy_os
        )

    def render_int_sersic(self, xc, yc, flux, r_eff, n, ellip, theta):
        im_no_os = render_sersic_2d(
            self.X, self.Y, xc, yc, flux, r_eff, n, ellip, theta
        )

        sub_im_os = render_sersic_2d(
            self.X_os, self.Y_os, xc, yc, flux, r_eff, n, ellip, theta
        )
        sub_im_os = jnp.sum(sub_im_os * self.w_os, axis=(2, 3))

        im = im_no_os.at[self.x_os_lo : self.x_os_hi, self.y_os_lo : self.y_os_hi].set(
            sub_im_os
        )
        return im

    def render_sersic(self, params: dict) -> jax.numpy.array:
        """Render a sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        im_int = self.render_int_sersic(
            params["xc"],
            params["yc"],
            params["flux"],
            params["r_eff"],
            params["n"],
            params["ellip"],
            params["theta"],
        )
        return self.fft_zeros, im_int, self.img_zeros

    def render_pointsource(self, params: dict) -> jax.numpy.array:
        """Render a Point source by interpolating given PSF into image. Currently jax only supports linear intepolation.

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        dx = params["xc"] - self.psf_shape[0] / 2.0
        dy = params["yc"] - self.psf_shape[1] / 2.0

        shifted_psf = jax.scipy.ndimage.map_coordinates(
            self.pixel_PSF * params["flux"],
            [self.X - dx, self.Y - dy],
            order=1,
            mode="constant",
        )

        return self.fft_zeros, self.img_zeros, shifted_psf


class FourierRenderer(BaseRenderer):
    """
    Class to render sources based on rendering them in Fourier space. Sersic profiles are modeled as a series of Gaussian following Shajib (2019) (https://arxiv.org/abs/1906.08263) and the implementation in lenstronomy (https://github.com/lenstronomy/lenstronomy/blob/main/lenstronomy/LensModel/Profiles/gauss_decomposition.py)
    """

    frac_start: float = eqx.field(static=True)
    frac_end: float = eqx.field(static=True)
    n_sigma: int = eqx.field(static=True)
    precision: int = eqx.field(static=True)
    use_interp_amps: bool = eqx.field(static=True)

    etas: jax.numpy.array
    betas: jax.numpy.array
    n_ax: jax.numpy.array
    amps_n_ax: jax.numpy.array

    def __init__(
        self,
        im_shape: Iterable,
        pixel_PSF: jax.numpy.array,
        frac_start: Optional[float] = 1e-2,
        frac_end: Optional[float] = 15.0,
        n_sigma: Optional[int] = 15,
        precision: Optional[int] = 10,
        use_interp_amps: Optional[bool] = True,
    ) -> None:
        """Initialize a Fourier renderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        frac_start : Optional[float], optional
            Fraction of r_eff for the smallest Gaussian component, by default 0.01
        frac_end : Optional[float], optional
            Fraction of r_eff for the largest Gaussian component, by default 15.
        n_sigma : Optional[int], optional
            Number of Gaussian Components, by default 15
        precision : Optional[int], optional
            precision value used in calculating Gaussian components, see Shajib (2019) for more details, by default 10
        use_interp_amps: Optional[bool]
            If True, instead of performing the direct calculation in Shajib (2019) at each iteration, a polynomial approximation is fit and used. The amplitudes of each gaussian component amplitudes as a function of Sersic index are interpolated based on a computed grid. This is much more numerically stable owing to the smooth gradients. If this matters for you then set this to False and make sure to enable jax's 64 bit capabilities which we find helps the stability.
        """
        super().__init__(im_shape, pixel_PSF)
        self.frac_start = frac_start
        self.frac_end = frac_end
        self.n_sigma = n_sigma
        self.precision = precision
        self.use_interp_amps = use_interp_amps
        self.etas, self.betas = calculate_etas_betas(self.precision)
        if not use_interp_amps and not jax.config.x64_enabled:
            warnings.warn(
                "!! WARNING !! - Gaussian decomposition can be numerically unstable when using jax's default 32 bit. Please either enable jax 64 bit or set 'use_poly_amps' = True in the renderer kwargs"
            )

        # Fit polynomial for smooth interpolation
        self.n_ax = jnp.linspace(0.65, 8.0, num=50)
        self.amps_n_ax = jax.vmap(
            lambda n: sersic_gauss_decomp(
                1.0,
                1.0,
                n,
                self.etas,
                self.betas,
                self.frac_start,
                self.frac_end,
                self.n_sigma,
            )[0]
        )(self.n_ax)

    def get_amps_sigmas(self, flux, r_eff, n):
        if self.use_interp_amps:
            amps_norm = interp1d(n, self.n_ax, self.amps_n_ax, method="cubic2")
            amps = amps_norm * flux
            sigmas = jnp.logspace(
                jnp.log10(r_eff * self.frac_start),
                jnp.log10(r_eff * self.frac_end),
                num=self.n_sigma,
            )
        else:
            amps, sigmas = sersic_gauss_decomp(
                flux,
                r_eff,
                n,
                self.etas,
                self.betas,
                self.frac_start * r_eff,
                self.frac_end * r_eff,
                self.n_sigma,
            )

        return amps, sigmas

    def render_sersic_mog_fourier(self, xc, yc, flux, r_eff, n, ellip, theta):
        amps, sigmas = self.get_amps_sigmas(flux, r_eff, n)
        q = 1.0 - ellip
        Fgal = render_gaussian_fourier(self.FX, self.FY, amps, sigmas, xc, yc, theta, q)
        return Fgal

    def render_sersic(self, params: dict) -> jax.numpy.array:
        """Render a Sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        F_im = self.render_sersic_mog_fourier(
            params["xc"],
            params["yc"],
            params["flux"],
            params["r_eff"],
            params["n"],
            params["ellip"],
            params["theta"],
        )
        return F_im, self.img_zeros, self.img_zeros

    def render_pointsource(self, params: dict) -> jax.numpy.array:
        """Render a Point source

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        F_im = render_pointsource_fourier(
            self.FX, self.FY, params["xc"], params["yc"], params["flux"]
        )
        return F_im, self.img_zeros, self.img_zeros


class HybridRenderer(FourierRenderer):
    """
    Class to render sources based on the hybrid rendering scheme introduced in Lang (2020). This avoids some of the artifacts introduced by rendering sources purely in Fourier space. Sersic profiles are modeled as a series of Gaussian following Shajib (2019) (https://arxiv.org/abs/1906.08263) and the implementation in lenstronomy (https://github.com/lenstronomy/lenstronomy/blob/main/lenstronomy/LensModel/Profiles/gauss_decomposition.py).

    Our scheme is implemented slightly differently than Lang (2020), specifically in how it chooses which gaussian components to render in Fourier vs. Real space. Lang (2020) employs a cutoff based on distance to the edge of the image. However given some of jax's limitation with dynamic shapes (see more here -> https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#dynamic-shapes), we have not implemented that specific criterion. Instead we use a simpler version where the user must decide how many components to render in real space, starting from the largest ones. While this is not ideal in all circumstances it still overcomes many of the issues of rendering purely in fourier space discussed in Lang (2020).
    """

    num_pixel_render: int = eqx.field(static=True)
    w_real: jax.numpy.array = eqx.field(static=True)
    w_fourier: jax.numpy.array = eqx.field(static=True)
    sig_psf_approx: float = eqx.field(static=True)

    def __init__(
        self,
        im_shape: Iterable,
        pixel_PSF: jax.numpy.array,
        frac_start: Optional[float] = 1e-2,
        frac_end: Optional[float] = 15.0,
        n_sigma: Optional[int] = 15,
        num_pixel_render: Optional[int] = 3,
        precision: Optional[int] = 10,
        use_interp_amps: Optional[bool] = True,
    ) -> None:
        """Initialize a  HybridRenderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        frac_start : Optional[float], optional
            Fraction of r_eff for the smallest Gaussian component, by default 0.02
        frac_end : Optional[float], optional
            Fraction of r_eff for the largest Gaussian component, by default 15.
        n_sigma : Optional[int], optional
            Number of Gaussian Components, by default 13
        num_pixel_Render : Optional[int], optional
            Number of components to render in pixel space, counts back from largest component
        precision : Optional[int], optional
            precision value used in calculating Gaussian components, see Shajib (2019) for more details, by default 10
        use_interp_amps: Optional[bool]
            If True, instead of performing the direct calculation in Shajib (2019) at each iteration, a polynomial approximation is fit and used. The amplitudes of each gaussian component amplitudes as a function of Sersic index are interpolated based on a computed grid. This is much more numerically stable owing to the smooth gradients. If this matters for you then set this to False and make sure to enable jax's 64 bit capabilities which we find helps the stability.
        """
        super().__init__(
            im_shape,
            pixel_PSF,
            frac_start,
            frac_end,
            n_sigma,
            precision,
            use_interp_amps,
        )

        self.num_pixel_render = num_pixel_render
        self.w_real = jnp.arange(
            self.n_sigma - self.num_pixel_render, self.n_sigma, dtype=jnp.int32
        )
        self.w_fourier = jnp.arange(
            self.n_sigma - self.num_pixel_render, dtype=jnp.int32
        )

        psf_X, psf_Y = jnp.meshgrid(
            jnp.arange(self.psf_shape[0]), jnp.arange(self.psf_shape[1])
        )
        sig_x = jnp.sqrt(
            (self.pixel_PSF * (psf_X - psf_X.mean()) ** 2).sum() / self.pixel_PSF.sum()
        )
        sig_y = jnp.sqrt(
            (self.pixel_PSF * (psf_Y - psf_Y.mean()) ** 2).sum() / self.pixel_PSF.sum()
        )
        self.sig_psf_approx = 0.5 * (sig_x + sig_y)

    def render_sersic_hybrid(self, xc, yc, flux, r_eff, n, ellip, theta):
        amps, sigmas = self.get_amps_sigmas(flux, r_eff, n)

        q = 1.0 - ellip

        sigmas_obs = jnp.sqrt(sigmas**2 + self.sig_psf_approx**2)
        q_obs = jnp.sqrt((q * q * sigmas**2 + self.sig_psf_approx**2) / sigmas_obs**2)

        Fgal = render_gaussian_fourier(
            self.FX,
            self.FY,
            amps[self.w_fourier],
            sigmas[self.w_fourier],
            xc,
            yc,
            theta,
            q,
        )

        im_gal = render_gaussian_pixel(
            self.X,
            self.Y,
            amps[self.w_real],
            sigmas_obs[self.w_real],
            xc,
            yc,
            theta,
            q_obs[self.w_real],
        )

        return Fgal, im_gal

    def render_sersic(self, params: dict) -> jax.numpy.array:
        """Render a Sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        F, im = self.render_sersic_hybrid(
            params["xc"],
            params["yc"],
            params["flux"],
            params["r_eff"],
            params["n"],
            params["ellip"],
            params["theta"],
        )
        return F, self.img_zeros, im

    def render_pointsource(self, params: dict) -> jax.numpy.array:
        """Render a Point source

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        F_im = render_pointsource_fourier(
            self.FX, self.FY, params["xc"], params["yc"], params["flux"]
        )
        return F_im, self.img_zeros, self.img_zeros


def sersic1D(
    r: Union[float, jax.numpy.array], flux: float, re: float, n: float
) -> Union[float, jax.numpy.array]:
    """Evaluate a 1D sersic profile

    Parameters
    ----------
    r : float
        radii to evaluate profile at
    flux : float
        Total flux
    re : float
        Effective radius
    n : float
        Sersic index

    Returns
    -------
    jax.numpy.array
        Sersic profile evaluated at r
    """
    bn = 1.9992 * n - 0.3271
    Ie = (
        flux
        / (re * re * 2 * jnp.pi * n * jnp.exp(bn + jax.scipy.special.gammaln(2 * n)))
        * bn ** (2 * n)
    )
    return Ie * jnp.exp(-bn * ((r / re) ** (1.0 / n) - 1.0))


def render_gaussian_fourier(
    FX: jax.numpy.array,
    FY: jax.numpy.array,
    amps: jax.numpy.array,
    sigmas: jax.numpy.array,
    xc: float,
    yc: float,
    theta: float,
    q: float,
) -> jax.numpy.array:
    """Render Gaussian components in the Fourier domain

    Parameters
    ----------
    FX : jax.numpy.array
        X frequency positions to evaluate
    FY : jax.numpy.array
        Y frequency positions to evaluate
    amps : jax.numpy.array
        Amplitudes of each component
    sigmas : jax.numpy.array
        widths of each component
    xc : float
        Central x position
    yc : float
        Central y position
    theta : float
        position angle
    q : float
        Axis ratio

    Returns
    -------
    jax.numpy.array
        Sum of components evaluated at FX and FY
    """
    theta = theta + (jnp.pi / 2.0)
    Ui = FX * jnp.cos(theta) + FY * jnp.sin(theta)
    Vi = -1 * FX * jnp.sin(theta) + FY * jnp.cos(theta)

    in_exp = (
        -1
        * (Ui * Ui + Vi * Vi * q * q)
        * (2 * jnp.pi * jnp.pi * sigmas * sigmas)[:, jnp.newaxis, jnp.newaxis]
        - 1j * 2 * jnp.pi * FX * xc
        - 1j * 2 * jnp.pi * FY * yc
    )
    Fgal_comp = amps[:, jnp.newaxis, jnp.newaxis] * jnp.exp(in_exp)
    Fgal = jnp.sum(Fgal_comp, axis=0)
    return Fgal


def render_pointsource_fourier(
    FX: jax.numpy.array, FY: jax.numpy.array, xc: float, yc: float, flux: float
) -> jax.numpy.array:
    """Render a point source in the Fourier domain

    Parameters
    ----------
    FX : jax.numpy.array
        X frequency positions to evaluate
    FY : jax.numpy.array
        Y frequency positions to evaluate
    xc : float
        Central x position
    yc : float
        Central y position
    flux : float
        Total flux of source

    Returns
    -------
    jax.numpy.array
        Point source evaluated at FX FY
    """
    in_exp = -1j * 2 * jnp.pi * FX * xc - 1j * 2 * jnp.pi * FY * yc
    F_im = flux * jnp.exp(in_exp)
    return F_im


def render_gaussian_pixel(
    X: jax.numpy.array,
    Y: jax.numpy.array,
    amps: jax.numpy.array,
    sigmas: jax.numpy.array,
    xc: float,
    yc: float,
    theta: float,
    q: Union[float, jax.numpy.array],
) -> jax.numpy.array:
    """Render Gaussian components in pixel space

    Parameters
    ----------
    FX : jax.numpy.array
        X positions to evaluate
    FY : jax.numpy.array
        Y positions to evaluate
    amps : jax.numpy.array
        Amplitudes of each component
    sigmas : jax.numpy.array
        widths of each component
    xc : float
        Central x position
    yc : float
        Central y position
    theta : float
        position angle
    q : Union[float,jax.numpy.array]
        Axis ratio

    Returns
    -------
    jax.numpy.array
        Sum of components evaluated at X and Y
    """
    X_bar = X - xc
    Y_bar = Y - yc
    theta = theta + (jnp.pi / 2.0)
    Xi = X_bar * jnp.cos(theta) + Y_bar * jnp.sin(theta)
    Yi = -1 * X_bar * jnp.sin(theta) + Y_bar * jnp.cos(theta)

    in_exp = (
        -1
        * (Xi * Xi + Yi * Yi / (q * q)[:, jnp.newaxis, jnp.newaxis])
        / (2 * sigmas * sigmas)[:, jnp.newaxis, jnp.newaxis]
    )
    im_comp = (amps / (2 * jnp.pi * sigmas * sigmas * q))[
        :, jnp.newaxis, jnp.newaxis
    ] * jnp.exp(in_exp)
    im = jnp.sum(im_comp, axis=0)
    return im


def render_sersic_2d(
    X: jax.numpy.array,
    Y: jax.numpy.array,
    xc: float,
    yc: float,
    flux: float,
    r_eff: float,
    n: float,
    ellip: float,
    theta: float,
) -> jax.numpy.array:
    """Evalulate a 2D Sersic distribution at given locations

    Parameters
    ----------
    X : jax.numpy.array
        x locations to evaluate at
    Y : jax.numpy.array
        y locations to evaluate at
    xc : float
        Central x position
    yc : float
        Central y position
    flux : float
        Total flux
    r_eff : float
        Effective radius
    n : float
        Sersic index
    ellip : float
        Ellipticity
    theta : float
        Position angle in radians [now measured from north]

    Returns
    -------
    jax.numpy.array
        Sersic model evaluated at given locations
    """
    bn = 1.9992 * n - 0.3271
    a, b = r_eff, (1 - ellip) * r_eff
    theta = theta + (jnp.pi / 2.0)
    cos_theta, sin_theta = jnp.cos(theta), jnp.sin(theta)
    x_maj = (X - xc) * cos_theta + (Y - yc) * sin_theta
    x_min = -(X - xc) * sin_theta + (Y - yc) * cos_theta
    amplitude = (
        flux
        * bn ** (2 * n)
        / (jnp.exp(bn + jax.scipy.special.gammaln(2 * n)) * r_eff**2 * jnp.pi * 2 * n)
    )
    z = jnp.sqrt((x_maj / a) ** 2 + (x_min / b) ** 2)
    out = amplitude * jnp.exp(-bn * (z ** (1 / n) - 1)) / (1.0 - ellip)
    return out


def calculate_etas_betas(precision: int) -> Tuple[jax.numpy.array, jax.numpy.array]:
    """Calculate the weights and nodes for the Gaussian decomposition described in Shajib (2019) (https://arxiv.org/abs/1906.08263)

    Parameters
    ----------
    precision : int
        Precision, higher number implies more precise decomposition but more nodes. Effective upper limit is 12 for 32 bit numbers, 27 for 64 bit numbers.

    Returns
    -------
    Tuple[jax.numpy.array, jax.numpy.array]
        etas and betas array to be use in gaussian decomposition
    """
    kes = jnp.arange(2 * precision + 1)
    betas = jnp.sqrt(2 * precision * jnp.log(10) / 3.0 + 2.0 * 1j * jnp.pi * kes)
    epsilons = jnp.zeros(2 * precision + 1)

    epsilons = epsilons.at[0].set(0.5)
    epsilons = epsilons.at[1 : precision + 1].set(1.0)
    epsilons = epsilons.at[-1].set(1 / 2.0**precision)

    for k in range(1, precision):
        epsilons = epsilons.at[2 * precision - k].set(
            epsilons[2 * precision - k + 1] + 1 / 2.0**precision * comb(precision, k)
        )

    etas = jnp.array(
        (-1.0) ** kes
        * epsilons
        * 10.0 ** (precision / 3.0)
        * 2.0
        * jnp.sqrt(2 * jnp.pi)
    )
    betas = jnp.array(betas)
    return etas, betas


def sersic_gauss_decomp(
    flux: float,
    re: float,
    n: float,
    etas: jax.numpy.array,
    betas: jax.numpy.array,
    sigma_start: float,
    sigma_end: float,
    n_comp: int,
) -> Tuple[jax.numpy.array, jax.numpy.array]:
    """Calculate a gaussian decomposition of a given sersic profile, following Shajib (2019) (https://arxiv.org/abs/1906.08263)

    Parameters
    ----------
    flux : float
        Total flux
    re : float
        half light radius
    n : float
        Sersic index
    etas : jax.numpy.array
        Weights for decomposition, can be calcualted using pysersic.rendering_utils.calculate_etas_betas
    betas : jax.numpy.array
        Nodes for decomposition, can be calcualted using pysersic.rendering_utils.calculate_etas_betas
    sigma_start : float
        width for the smallest Gaussian component
    sigma_end : float
        width for the largest Gaussian component
    n_comp : int
        Number of Gaussian components

    Returns
    -------
    Tuple[jax.numpy.array, jax.numpy.array]
        Amplitudes and sigmas of Gaussian decomposition
    """
    sigmas = jnp.logspace(jnp.log10(sigma_start), jnp.log10(sigma_end), num=n_comp)

    f_sigmas = jnp.sum(
        etas * sersic1D(jnp.outer(sigmas, betas), flux, re, n).real, axis=1
    )

    del_log_sigma = jnp.abs(jnp.diff(jnp.log(sigmas)).mean())

    amps = f_sigmas * del_log_sigma / jnp.sqrt(2 * jnp.pi)

    amps = amps.at[0].multiply(0.5)
    amps = amps.at[-1].multiply(0.5)

    amps = amps * 2 * jnp.pi * sigmas * sigmas

    return amps, sigmas



import jax.numpy as jnp
from numpyro import distributions as dist, sample, handlers, factor,deterministic
from typing import Optional

def gaussian_loss(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                suffix: str = "")-> float:
    """Basic Gaussian loss function using given uncertainties

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """

    with handlers.mask(mask = mask):
        loss = sample(f"Loss{suffix}", dist.Normal(mod, rms), obs=data)
    return loss

def cash_loss(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                suffix: str = "")-> float:
    """
    Cash statistic based on Poisson statistics derived in Cash (1979) (DOI 10.1086/156922) and advocated for in Erwin (2015) (https://arxiv.org/abs/1408.1097) for use in Sersic fitting. Since the is based on Poisson statistics, scaling of the image will produce different confidence intervals. Additionally, since a logarithm is taken of the model image, negative values associated with different sky models will cause issues.

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    
    with handlers.mask(mask = mask):
        loss = factor(f'cash_loss{suffix}', -1*(mod - data*jnp.log(mod)))
    return loss

def gaussian_loss_w_frac(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                suffix: str = "")-> float:
    """Gaussian loss with and additional fractional increase to all uncertainties such that,

    $$ \sigma_{new,i} = (1 + f) * \sigma_{old,i} $$

    f is a free parameter varied between -0.5 and 2

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    
    scatter_frac = sample(f'frac_rms_increase{suffix}', dist.TruncatedNormal(low = -0.5, high = 2) )
    with handlers.mask(mask = mask):
        loss =  sample(f"Loss{suffix}", dist.Normal(mod, (1+ scatter_frac)*rms), obs=data)    
    return loss

def gaussian_loss_w_sys(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                suffix: str = "")-> float:
    """Gaussian loss with and additional systematic increase such_that

    $$ \sigma_{new,i}^2 = \sigma_{old,i}^2 + \sigma_{sys}^2 $$

    \sigma_{sys} is a free parameter with a Normal prior, with loc = 0 and scale = mean(rms) truncated to ensure > 0

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    
    sys_scatter_base = sample(f'sys_rms_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1 ) )
    sys_scatter = deterministic(f'sys_rms{suffix}', sys_scatter_base*jnp.mean(rms))
    with handlers.mask(mask = mask):
        loss =  sample(f"Loss{suffix}", dist.Normal(mod, jnp.sqrt(rms**2 + sys_scatter**2)), obs=data)    
    return loss

def student_t_loss(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                nu: Optional[int] = 5,
                suffix: str = "")-> float:
    """
    Student T loss, with a df = 5 by default. This has fatter tails than Gaussian loss (or chi squared) so is more resilient to outliers

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    nu = 5.
    with handlers.mask(mask = mask):
        loss =  sample(f"Loss{suffix}", dist.StudentT(nu,mod,jnp.sqrt((nu-2.)/2.)*rms), obs=data)    
    return loss

def student_t_loss_free_sys(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                nu: Optional[int] = 5,
                suffix: str = "")-> float:
    """
    Student T loss, which has fatter tails than Gaussian loss (or chi squared) so is so is more resilient to outliers. In addition, add additional systematic increase such that

    $$ \sigma_{new,i}^2 = \sigma_{old,i}^2 + \sigma_{sys}^2 $$

    \sigma_{sys} is a free parameter with a Normal prior, with loc = 0 and scale = mean(rms) truncated to ensure > 0


    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    sys_scatter_base = sample(f'sys_rms_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1 ) )
    sys_scatter = deterministic(f'sys_rms{suffix}', sys_scatter_base*jnp.mean(rms))
    rms_new = jnp.sqrt((nu-2.)/2.)*jnp.sqrt(rms**2 + sys_scatter**2)

    with handlers.mask(mask = mask):
        loss =  sample(f"Loss{suffix}", dist.StudentT(nu, mod, rms_new), obs=data)    
    return loss

def pseudo_huber_loss(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                delta: Optional[int] = 3,
                suffix: str = "")-> float:
    """
    Pseudo huber loss function of the form:

    $$ L = \delta^2 * ( \sqrt{1 + (a/\delta)^2} - 1) $$

    where a is the residuals scaled by the rms and delta can be chosen. This loss function is more robust to outliers than the Gaussian loss function as it is meant to transition for L2 to L1 loss at residuals greater than delta. The delta parameter is 3 by default but can be varied.

    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties

    Returns
    -------
    float
        Sampled loss function
    """
    
    with handlers.mask(mask = mask):
        res = (data-mod)/rms
        loss = factor(f'pseudo_huber_loss{suffix}', -1.*(jnp.sqrt(1 + ( res/delta )**2) - 1) )
    return loss

def gaussian_mixture(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                c: Optional[float] = 5.,
                suffix: str = "")-> float:
    """
    Gaussian mixture loss function, with one representing a "contaminating" outlier distribution with standard deviation equal to c*rms where c is 5 by default. The "contaminating fraction" or fraction of outliers is a free parameter with a Uniform prior between 0 and 0.25.

    Parameters
    ----------
    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties
    c : float, optional
        factor to increase rms for outlier distribution, by default 5

    Returns
    -------
    float
        _description_
    """
    contam_frac_base = sample(f'outlier_frac_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1., high = 5.) )
    contam_frac = deterministic(f'outlier_frac{suffix}', contam_frac_base*0.05 )

    mixture_dists = dist.Categorical(probs = jnp.array([1-contam_frac, contam_frac]))
    component_dists = dist.Normal(jnp.stack([mod,mod],axis = -1), jnp.stack([rms,c*rms],axis = -1) )

    with handlers.mask(mask = mask):
        loss = sample(f"Loss{suffix}", dist.MixtureSameFamily(mixture_dists, component_dists), obs=data)
    return loss

def gaussian_mixture_w_sys(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                c: Optional[float] = 5.,
                suffix: str = "")-> float:
    """
    Gaussian mixture loss function, with one representing a "contaminating" outlier distribution with standard deviation equal to c*rms where c is 5 by default. The "outlier fraction" or fraction of outliers is a free parameter with a Uniform prior between 0 and 0.25.

    Parameters
    ----------
    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties
    c : float, optional
        factor to increase rms for outlier distribution, by default 5

    Returns
    -------
    float
        _description_
    """
    contam_frac_base = sample(f'outlier_frac_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1., high = 5.) )
    contam_frac = deterministic(f'outlier_frac{suffix}', contam_frac_base*0.05 )

    sys_scatter_base = sample(f'sys_rms_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1 ) )
    sys_scatter = deterministic(f'sys_rms{suffix}', sys_scatter_base*jnp.mean(rms))

    rms_new = jnp.sqrt(rms**2 + sys_scatter**2)
    mixture_dists = dist.Categorical(probs = jnp.array([1-contam_frac, contam_frac]))
    component_dists = dist.Normal(jnp.stack([mod,mod],axis = -1), jnp.stack([rms_new,c*rms_new],axis = -1) )

    with handlers.mask(mask = mask):
        loss = sample(f"Loss{suffix}", dist.MixtureSameFamily(mixture_dists, component_dists), obs=data)
    return loss

def gaussian_mixture_w_frac(mod: jnp.array,
                data: jnp.array,
                rms: jnp.array,
                mask: jnp.array,
                c: Optional[float] = 5.,
                suffix: str = "")-> float:
    """
    Gaussian mixture loss function, with one representing a "contaminating" outlier distribution with standard deviation equal to c*rms where c is 5 by default. The "outlier fraction" or fraction of outliers is a free parameter with a Uniform prior between 0 and 0.25.

    Parameters
    ----------
    Parameters
    ----------
    mod : jnp.array
        Model image
    data : jnp.array
        data to be fit
    rms : jnp.array
        per pixel 1-sigma uncertainties
    c : float, optional
        factor to increase rms for outlier distribution, by default 5

    Returns
    -------
    float
        _description_
    """
    contam_frac_base = sample(f'outlier_frac_base{suffix}', dist.TruncatedNormal(low = 0, scale = 1., high = 5.) )
    contam_frac = deterministic(f'outlier_frac{suffix}', contam_frac_base*0.05 )

    sig_frac = sample(f'rms_frac{suffix}', dist.TruncatedNormal(low = -2./3., high = 2., loc = 0, scale = 0.5 ) )

    rms_new = (1+sig_frac)*rms
    mixture_dists = dist.Categorical(probs = jnp.array([1-contam_frac, contam_frac]))
    component_dists = dist.Normal(jnp.stack([mod,mod],axis = -1), jnp.stack([rms_new,c*rms],axis = -1) )

    with handlers.mask(mask = mask):
        loss = sample(f"Loss{suffix}", dist.MixtureSameFamily(mixture_dists, component_dists), obs=data)
    return loss



import pytest
import numpy as np
import copy
from pysersic import FitSingle,FitMulti, priors, rendering
from astropy.convolution import Gaussian2DKernel
from jax.random import PRNGKey
import arviz
from numpyro import distributions as dist,infer,sample, optim
from pysersic.pysersic import train_numpyro_svi_early_stop
from pysersic.priors import estimate_sky
from pysersic.multiband import FitMultiBandBSpline, FitMultiBandPoly



rng = np.random.default_rng(seed=10)
psf = Gaussian2DKernel(x_stddev=2.5).array
renderer = rendering.HybridRenderer((40,40), psf)

band_list = ['a','b','c']
wv_list = np.array([0.5,1,1.5])
linked_params = ['flux',]
const_params = ['xc',]
wv_to_save = np.linspace(0.5,5, num = 10)

im_list = []
rms_list = []
for wv in wv_list:
    rms = np.ones((40,40))*0.05*wv
    im_s = np.zeros((40,40))
    im_s += rng.normal(scale = 0.05*wv, size = (40,40))
    im_s = im_s + renderer.render_source( dict(xc = 20.,yc=20., flux = 200.*wv), 'pointsource' )
    im_list.append(im_s)
    rms_list.append(rms)

prior = priors.PySersicSourcePrior('pointsource')
prior.set_gaussian_prior('xc', 20.,1.)
prior.set_gaussian_prior('yc', 20.,1.)
prior.set_uniform_prior('flux', 25.,400.)

fitter_list = []

for j in range(3):
    fitter_list.append(FitSingle(
        data=im_list[j], 
        rms = rms_list[j],
        psf = psf,
        prior = copy.deepcopy(prior), 
    ) )

@pytest.mark.parametrize('fitter', [FitMultiBandPoly, FitMultiBandBSpline] )
def test_multiband(fitter):
    mb_fitter = fitter(fitter_list=fitter_list,
                            wavelengths=wv_list,
                            band_names= band_list,
                            linked_params=linked_params,
                            const_params=const_params,
                            wv_to_save= wv_to_save)

    #Make sure everything was parsed correctly
    assert 'flux' in mb_fitter.linked_params
    assert 'yc' in  mb_fitter.unlinked_params
    assert 'xc' in  mb_fitter.const_params

    res = mb_fitter.estimate_posterior(rkey = PRNGKey(11))
    summ_dict = res.summary().to_dict(orient = 'index')

    #Make sure all of the variables that should exist do
    assert 'flux_at_wv[9]' in summ_dict
    assert 'flux_c' in summ_dict

    assert 'yc_c' in summ_dict
    assert 'xc' in summ_dict

    #Make sure some of the values are right to test if inference was successful
    assert summ_dict['flux_b']['mean'] == pytest.approx(200, rel = 5e-2)
    assert summ_dict['xc']['mean'] == pytest.approx(20, rel = 5e-2) 
    assert summ_dict['yc_b']['mean'] == pytest.approx(20, rel = 5e-2) 



from .pysersic import BaseFitter, FitMulti,FitSingle
from .priors import update_prior_suffix
import abc
import numpy as np
import jax.numpy as jnp
import copy
from numpyro import distributions as dist, sample, deterministic, handlers, infer
import jax

from scipy.interpolate import make_interp_spline
from typing import List,Union,Optional
import arviz as az


class BaseMultiBandFitter(BaseFitter):
    """
    Base class for multi-band fitter, for new classes only need to add a `sample_param_at_bands` function specifying the linking of parameters between bands.
    """
    def __init__(self, fitter_list: Union[List[FitSingle], List[FitMulti]],
                  wavelengths: jax.Array,
                  linked_params: List[str],
                  const_params: Optional[List[str]] = [],
                  band_names: Optional[List[str]] = None,
                  linked_params_range: Optional[dict] = {}, 
                  wv_to_save: Optional[jax.Array] = None,
                  rescale_unlinked_priors: Optional[bool] = False,) -> None:
        """Base class to multi-band fitting, to be used in other classes where the "linking" function is defined

        Parameters
        ----------
        fitter_list : List[FitSingle]
            List of FitSingle objects for all the different bands
        wavelengths : jax.Array
            The wavelengths of each band, They will be normalized so don't need to rescale
        linked_params : Optional[List[str]], optional
            Which parameters to link across bands, by default None
        const_params : Optional[List[str]], optional
            Which parameters to hold constant across bands, by default None
        band_names : Optional[List[str]], optional
            Names of bands, if None will default to 'band_0', 'band_1' etc.
        linked_params_range : Optional[dict], optional
            Ranges to enforce on any link parameters, by default {}
        wv_to_save : Optional[jax.Array], optional
            Additional wavelengths to track linking parameters for, by default None
        rescale_unlinked_priors: Optional[bool]:
            Whether or not to rescale priors of unlinked parameters based on Laplace posteriors from individual band fits
        """
        self.fitter_list = copy.deepcopy(fitter_list)

        self.n_bands = len(fitter_list)
        assert len(wavelengths) == self.n_bands, "Number of fitters given does not much length of wavelength array"
        self.wavelengths = jnp.array(wavelengths)
        
        for fitter in self.fitter_list:
            assert isinstance(fitter, type(fitter_list[0])), "All fitters must be of same type"
        
        if isinstance(self.fitter_list[0], FitSingle):
            self.fitter_type = 'single'
        else:
            self.fitter_type = 'multi'
 
        if band_names is None:
            self.band_names = [f'Band_{i}' for i in range(len(self.wavelengths))]
        else:
            self.band_names = band_names
    
        # Make sure all priors have the same number of parameters
        self.param_names = self.fitter_list[0].prior.param_names
        for band_fitter in self.fitter_list:
            assert band_fitter.prior.param_names  == self.param_names, "All fitters must have the same parameters"

        for band,fitter in zip(self.band_names,self.fitter_list):
            fitter.prior = update_prior_suffix(fitter.prior, f'_{band}')

        # Normalize lamdda's to help convergence
        self.wv_av = np.mean(self.wavelengths)
        self.wv_range = np.max(self.wavelengths) - np.min(self.wavelengths)
        self.wv_normed = (self.wavelengths - self.wv_av)/ self.wv_range
        self.wv_to_save = wv_to_save
        if self.wv_to_save is not None:
            self.wv_to_save_normed = (self.wv_to_save - self.wv_av)/self.wv_range
        else:
            self.wv_to_save_normed = None


        # Set up which parameters are linked and which are not
        self.linked_params = linked_params
        for param in self.linked_params:
            assert param in self.param_names, f"Linked parameter ({param}) not in parameter list"
        
        self.const_params = const_params
        for param in self.const_params:
            assert param in self.param_names, f"const parameter ({param}) not in parameter list"

        self.unlinked_params = [p for p in self.param_names if (p not in self.linked_params and p not in self.const_params)]

        #Update reparam dict, linked parameters are handled in each subclass
        self.reparam_dict = {}
        for band_name, band_fitter in zip(self.band_names, self.fitter_list):
            for param_name in self.unlinked_params:
                val = band_fitter.prior.reparam_dict[f'{param_name}_{band_name}']
                self.reparam_dict.update( {f'{param_name}_{band_name}':val} )

        #Find mean and scale of prior samples of all bands to re-scale
        self.linked_params_mean = {}
        self.linked_params_scale = {}
        key = jax.random.key(self.n_bands)
        for param in self.linked_params:
                params_prior_samples = []
                for fitter, band in zip(self.fitter_list, self.band_names):
                    key, _ = jax.random.split(key)
                    dist_cur = fitter.prior.dist_dict[f'{param}_{band}']
                    params_prior_samples.append( dist_cur.sample(key, sample_shape = (200,)) )
                params_prior_samples = jnp.stack(params_prior_samples)

                self.linked_params_mean[param] = params_prior_samples.mean()
                self.linked_params_scale[param] = params_prior_samples.std()


        #Set defaults for all parameters that have known limits
        self.linked_params_range = {}
        for param in self.linked_params:
            if 'n' in param:
                self.linked_params_range[param] = [0.65,8]
            if 'ellip' in param:
                self.linked_params_range[param] = [0.,0.9]
            if 'theta' in param:
                self.linked_params_range[param] = [0,2*np.pi]

        self.linked_params_range.update(linked_params_range) #User-specific updates


        #For constant parameters use the prior in the of the first band
        self.const_prior_dict = {}
        for const_param in self.const_params:
            param_in_band_name = f'{const_param}_{self.band_names[0]}'
            prior_dist = copy.copy(self.fitter_list[0].prior.dist_dict[param_in_band_name] )
            self.const_prior_dict[const_param] = copy.deepcopy(prior_dist)
            
            if param_in_band_name in self.fitter_list[0].prior.reparam_dict:
                self.reparam_dict[const_param] = self.fitter_list[0].prior.reparam_dict[param_in_band_name]
        

        # Use estimated posterior of each individual band to set priors on unlinked parameters
        # Since these are "nuisance" in some sense that should help convergence etc. when sampling
        if rescale_unlinked_priors:
            for fitter in self.fitter_list:
                if hasattr(fitter, 'svi_results'):

                        svi_summ_no_suffix = az.summary(fitter.svi_results.idata, round_to=5).T.to_dict()
                        svi_summ = {}
                        for key in svi_summ_no_suffix:
                            svi_summ[f'{key}{fitter.prior.suffix}'] = svi_summ_no_suffix[key]
                else:    
                    svi_res = fitter.estimate_posterior()
                    svi_summ = az.summary(svi_res.idata, round_to=5)

    
                    
                for key, summ_param in svi_summ.items():
                    param = key.replace(fitter.prior.suffix, '')
                    if param in self.unlinked_params:
                        prior_dist = fitter.prior.dist_dict[key]

                        #Check if prior dist is transformed, most likely yes
                        is_t = isinstance(prior_dist, dist.TransformedDistribution)
                        if is_t:
                            base_dist = prior_dist.base_dist
                        else:
                            base_dist = prior_dist
                        #Check if there are bounds on prior
                        if isinstance(base_dist.support,dist.constraints._Real):
                            fitter.prior.set_gaussian_prior(param, summ_param['mean'], summ_param['sd']*2.)
                        else:
                            #If so then use a truncated distribution
                            low = base_dist.support.lower_bound
                            if is_t:
                                for t in prior_dist.transforms:
                                    low = t(low)
                            try:
                                high = base_dist.support.upper_bound
                                for t in prior_dist.transforms:
                                    high = t(high)
                            except AttributeError:  # Sometimes distributions are only lower bound e.g. r_eff at 0.5 pixels
                                high = None
                                
                            fitter.prior.set_truncated_gaussian_prior(param, summ_param['mean'], summ_param['sd']*2., low = low, high = high)
                
                if fitter.prior.sky_type != 'none':
                    sky_params = [param for param in svi_summ.keys() if 'sky' in param]
                    for param in sky_params:
                        param_no_suffix = param.replace(fitter.prior.suffix, '')
                        fitter.prior.sky_prior.update_prior(param_no_suffix, svi_summ[param]['mean'], svi_summ[param]['sd']*2.)
        
        ##Throw all the data into a dictionary so it gets past onto results class if needed
        self.data = {b:fitter.data for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.rms = {b:fitter.rms for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.psf = {b:fitter.psf for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.mask = {b:fitter.mask for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.loss_func = {b:fitter.loss_func for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.renderer = {b:fitter.renderer for (b,fitter) in zip(self.band_names,self.fitter_list)}
        self.prior = {b:fitter.prior for (b,fitter) in zip(self.band_names,self.fitter_list)}
    
    @abc.abstractmethod
    def sample_param_at_bands(self, name: str) -> jax.Array:
        """Function used to sample linked parameters at each band

        Parameters
        ----------
        name : str
            parameter name

        Returns
        -------
        params_at_bands : jax.Array
            parameter values sampled at bands
        """
        return NotImplementedError

    def build_model(self,return_model : bool = False) -> callable:
        """build numpyro model for multi-band inference

        Parameters
        ----------
        return_model : bool, optional
            wether or not to save and return observed images, by default False

        Returns
        -------
        callable
            numpyro model
        """
        @handlers.reparam(config = self.reparam_dict)
        def model():
            #Draw heriarchichal parameters for each linked parameter
            params_dict = {}
            for name in self.band_names:
                params_dict[name] = {}
            
            for param_name in self.linked_params:
                x = self.sample_param_at_bands(param_name)
                for i,band_name in enumerate(self.band_names):
                    params_dict[band_name][param_name] = deterministic(f'{param_name}_{band_name}', x[i])

            for param, prior_dist in self.const_prior_dict.items():
                value = sample(param, prior_dist)
                for band in self.band_names:
                    params_dict[band][param] = value

            all_obs = []
            for band_name,band_fitter in zip(self.band_names, self.fitter_list):
                #Draw values for unlinked parameters based on individual priors (with suffix_names)
                for param_name in self.unlinked_params:
                    params_dict[band_name][param_name] = sample(f'{param_name}_{band_name}', band_fitter.prior.dist_dict[f'{param_name}_{band_name}'])
                
                #Render each band like normal
                if self.fitter_type == 'single':
                    out = band_fitter.renderer.render_source(params_dict[band_name],
                                                             band_fitter.prior.profile_type,
                                                             suffix = "")
                else:
                    out = band_fitter.renderer.render_for_model(params_dict[band_name],
                                                                band_fitter.prior.catalog['type'],
                                                                suffix = "")


                sky = band_fitter.prior.sample_sky(band_fitter.renderer.X, band_fitter.renderer.Y)
                obs = out + sky
                all_obs.append(obs)
                band_fitter.loss_func(obs, band_fitter.data, band_fitter.rms, band_fitter.mask, suffix = f'_{band_name}')
            if return_model:
                deterministic('model', jnp.array(all_obs))
        return model
    

class FitMultiBandPoly(BaseMultiBandFitter):
    def __init__(self, 
                 fitter_list: List[FitSingle],
                 wavelengths: jax.Array,
                 linked_params: List[str],
                 const_params: Optional[List[str]] = [],
                 band_names: Optional[List[str]] = None,
                 linked_params_range: Optional[dict] = {},
                 wv_to_save: Optional[jax.Array] = None,
                 rescale_unlinked_priors: Optional[bool] = False,
                 poly_order: Optional[int] = 2,) -> None:
        """
        Multi-band fitting for where a polynomial is used as the linking function for specified parameters. For parameters with bounds, the output of the polynomial function is passed through a logisitic function to impose constraints in a continuous and differentiable way.

        Parameters
        ----------
        fitter_list : List[FitSingle]
            List of FitSingle objects for all the different bands
        wavelengths : jax.Array
            The wavelengths of each band, They will be normalized so don't need to rescale
        linked_params : Optional[List[str]], optional
            Which parameters to link across bands, by default None
        const_params : Optional[List[str]], optional
            Which parameters to hold constant across bands, by default None
        band_names : Optional[List[str]], optional
            Names of bands, if None will default to 'band_0', 'band_1' etc.
        linked_params_range : Optional[dict], optional
            Ranges to enforce on any link parameters, by default {}
        wv_to_save : Optional[jax.Array], optional
            Additional wavelengths to track linking parameters for, by default None
        rescale_unlinked_priors: Optional[bool]:
            Whether or not to rescale priors of unlinked parameters based on Laplace posteriors from individual band fits
        poly_order: Optional[int]:
            Order of polynomial (i.e. linear = 1, quadratic = 2) for the linking function, default 2
        """

        super().__init__(fitter_list, wavelengths, linked_params,const_params, band_names, linked_params_range, wv_to_save, rescale_unlinked_priors=rescale_unlinked_priors)
        self.poly_order = poly_order + 1 # account for constant term, change to the more intuitive definition

    def restrict_func(self,x, hi,low):
        return  jax.lax.logistic(x) *(hi-low) + low

    def sample_param_at_bands(self, name):
        prior_sigma = 1.

        poly_coeff = sample(f'{name}_poly_coeff',dist.Normal(scale=prior_sigma), sample_shape=(self.poly_order,))
        value_at_bands_normed = jnp.polyval(poly_coeff, self.wv_normed)
        value_to_save_normed = jnp.polyval(poly_coeff, self.wv_to_save_normed)

        if name in self.linked_params_range:
            low,hi = self.linked_params_range[name]
            
            # Use Tanh to make sure parameters are constrained
            value_at_bands = self.restrict_func(value_at_bands_normed,hi,low)
            
            deterministic(f'{name}_at_wv', 
                            self.restrict_func(value_to_save_normed,hi,low)
            )
        else:
            value_at_bands = value_at_bands_normed*self.linked_params_scale[name] + self.linked_params_mean[name]

            deterministic(f'{name}_at_wv', 
                            value_to_save_normed*self.linked_params_scale[name] + self.linked_params_mean[name]
            )

        return value_at_bands


class FitMultiBandBSpline(BaseMultiBandFitter):
    def __init__(self, 
                 fitter_list: List[FitSingle],
                 wavelengths: jax.Array,
                 linked_params: List[str],
                 const_params: Optional[List[str]] = [],
                 band_names: Optional[List[str]] = None,
                 linked_params_range: Optional[dict] = {},
                 wv_to_save: Optional[jax.Array] = None,
                 rescale_unlinked_priors: Optional[bool] = False,
                 N_knots: Optional[int] = 4,
                 spline_k: Optional[int] = 2,
                 pad_knots: Optional[bool] = True,
                ) -> None:
        """
        Multi-band fitting for where a B-spline is used as the linking function for specified parameters. 

        Parameters
        ----------
        fitter_list : List[FitSingle]
            List of FitSingle objects for all the different bands
        wavelengths : jax.Array
            The wavelengths of each band, They will be normalized so don't need to rescale
        linked_params : Optional[List[str]], optional
            Which parameters to link across bands, by default None
        const_params : Optional[List[str]], optional
            Which parameters to hold constant across bands, by default None
        band_names : Optional[List[str]], optional
            Names of bands, if None will default to 'band_0', 'band_1' etc.
        linked_params_range : Optional[dict], optional
            Ranges to enforce on any link parameters, by default {}
        wv_to_save : Optional[jax.Array], optional
            Additional wavelengths to track linking parameters for, by default None
        rescale_unlinked_priors : Optional[bool]
            Whether or not to rescale priors of unlinked parameters based on Laplace posteriors from individual band fits
        N_knots : Optional[int]
            Number of spline knots, uniformly spaced, default is 4
        spline_k : Optional[int]
            B-spline order to use, default 2
        pad_knots : Optional[bool]
            Wether or not to pad location of first and last knots by +/- 10% of wv range from min and max of wavelengths 
        """
        super().__init__(fitter_list, wavelengths, linked_params,const_params, band_names, linked_params_range, wv_to_save, rescale_unlinked_priors=rescale_unlinked_priors)
        self.N_knots = N_knots
        self.spline_k = spline_k
        min_lambda, max_lambda = jnp.min(self.wavelengths), jnp.max(self.wavelengths)
        
        #Pad knots by 10% on either side, leads to more uniform "importance"
        if pad_knots:
            lambda_pad = (max_lambda - min_lambda)/10.
        else:
            lambda_pad = 0
        
        #Set up spline matrix
        bspl_class = make_interp_spline(x = np.linspace(min_lambda-lambda_pad, max_lambda + lambda_pad, num = self.N_knots, endpoint=True),
                                                y = np.ones(self.N_knots), k = spline_k)
        self.dmat_bands = jnp.array( bspl_class.design_matrix(self.wavelengths, bspl_class.t,k = spline_k).toarray() )
        wv_to_save_dmat = jnp.clip(self.wv_to_save, a_min = min_lambda, a_max=max_lambda)
        self.dmat_save = jnp.array( bspl_class.design_matrix(wv_to_save_dmat, bspl_class.t,k = spline_k).toarray() )
    
    def sample_param_at_bands(self, name):
        prior_sigma = 1.5
        m,s = self.linked_params_mean[name], self.linked_params_scale[name]

        if name in self.linked_params_range:
            low,hi = self.linked_params_range[name]
            
            dist_to_sample = dist.TransformedDistribution(
                dist.Uniform(low=jnp.zeros(self.N_knots), high = jnp.ones(self.N_knots)),
                dist.transforms.AffineTransform(scale=hi-low, loc = low)
            )

        else:
            dist_to_sample = dist.TransformedDistribution(
                dist.Normal(scale = prior_sigma*jnp.ones(self.N_knots) ),
                dist.transforms.AffineTransform(scale=s, loc = m)
            )

        with handlers.reparam(config={f'bspl_w_{name}': infer.reparam.TransformReparam()}):
            weights_cur = sample(f'bspl_w_{name}',dist_to_sample)
        
        value_at_bands = jnp.dot(self.dmat_bands,weights_cur)
        
        deterministic(f'{name}_at_wv', 
                        jnp.dot(self.dmat_save,weights_cur)
        )

        return value_at_bands




from pysersic.rendering import PixelRenderer,FourierRenderer,HybridRenderer
from pysersic.rendering import render_sersic_2d
from astropy.convolution import Gaussian2DKernel
import pytest
import jax.numpy as jnp
from scipy.integrate import dblquad
from functools import partial
import jax

render_sersic_2d = jax.jit(render_sersic_2d)
kern = Gaussian2DKernel(x_stddev= 1.5)
psf = jnp.array(kern.array)
err_tol = 0.01 # 1% error tolerence for total flux

pixel_renderer = PixelRenderer((150,150), psf)
fourier_renderer = FourierRenderer((150,150), psf)
hybrid_renderer = HybridRenderer((150,150), psf)

@partial(jax.jit, static_argnums = [1,])
def get_models(params, profile_type: str):
    im_px = pixel_renderer.render_source(params, profile_type=profile_type)
    im_fr = fourier_renderer.render_source(params, profile_type=profile_type)
    im_hy = hybrid_renderer.render_source(params, profile_type=profile_type)
    return im_px, im_fr,im_hy

@pytest.mark.parametrize("pos", [(75.,75.),(75.5,75.5),(75.25,75.25) ,(75.,75.5),(75.5,75.)]) 
def test_point_source(pos):
    flux = 10.
    params = dict(flux = flux, xc = pos[0], yc = pos[1] )
    ims = get_models(params, 'pointsource')
    assert pytest.approx(ims[0].sum(), rel = err_tol) == flux #pixel
    assert pytest.approx(ims[1].sum(), rel = err_tol) == flux #fourier
    assert pytest.approx(ims[2].sum(), rel = err_tol) == flux #hybrid

@pytest.mark.parametrize("pos", [(75.,75.),(75.5,75.5),])
@pytest.mark.parametrize("re", [3.,5.]) 
@pytest.mark.parametrize("n", [1.5,2.5]) 
@pytest.mark.parametrize("ellip", [0,0.5])
@pytest.mark.parametrize("theta", [0,3.14/4.]) 
def test_sersic(pos,re,n,ellip,theta):
    flux = 10
    params = dict(flux = flux, xc = pos[0], yc = pos[1], n = n, ellip = ellip, theta = theta, r_eff = re )
    #Calculate fraction of flux contained in image
    int_test = partial(render_sersic_2d, xc = pos[0],yc = pos[1], flux = 10, r_eff = re, n = n, ellip = ellip,theta = theta)
    to_int = lambda x,y: float(int_test(x,y))
    lo_fun = lambda x: 0.
    hi_fun = lambda x: 150. 
    flux_int,_ = dblquad(to_int, 0.,150., lo_fun,hi_fun,epsrel=5.e-3)
    
    ims = get_models(params, 'sersic')
    assert pytest.approx(float(ims[0].sum()), rel = err_tol) == flux_int #pixel
    assert pytest.approx(float(ims[1].sum()), rel = err_tol) == flux_int #fourier
    assert pytest.approx(float(ims[2].sum()), rel = err_tol) == flux_int #hybrid

@pytest.mark.parametrize("prof", ['exp','dev'])
@pytest.mark.parametrize("pos", [(75.,75.),(75.5,75.5),])
@pytest.mark.parametrize("re", [3.,5.]) 
@pytest.mark.parametrize("ellip", [0,0.5])
@pytest.mark.parametrize("theta", [0,3.14/4.]) 
def test_exp_dev(prof,pos,re,ellip,theta):
    flux = 10.
    params = dict(flux = flux, xc = pos[0], yc = pos[1], ellip = ellip, theta = theta, r_eff = re )
    
    if prof == 'exp':
        n=1.
    if prof == 'dev':
        n=4.

    #Calculate fraction of flux contained in image
    int_test = partial(render_sersic_2d, xc = pos[0],yc = pos[1], flux = 10, r_eff = re, n = n, ellip = ellip,theta = theta)
    to_int = lambda x,y: float(int_test(x,y))

    lo_fun = lambda x: 0.
    hi_fun = lambda x: 150. 
    flux_int,_ = dblquad(to_int, 0.,150., lo_fun,hi_fun,epsrel=5.e-3)

    ims = get_models(params, prof)
    assert pytest.approx(float(ims[0].sum()), rel = err_tol) == flux_int #pixel
    assert pytest.approx(float(ims[1].sum()), rel = err_tol) == flux_int #fourier
    assert pytest.approx(float(ims[2].sum()), rel = err_tol) == flux_int #hybrid



import pytest
import jax.numpy as jnp
import numpy as np
from pysersic.priors import BasePrior, PySersicMultiPrior, PySersicSourcePrior, SourceProperties, estimate_sky, NoSkyPrior,FlatSkyPrior, TiltedPlaneSkyPrior
from numpyro.handlers import seed
from numpyro import distributions as dist
 
prof_names = ['sersic','doublesersic','pointsource','exp','dev']
prof_vars = [ ['xc','yc','flux','r_eff','n','ellip','theta'],
        ['xc','yc','flux','f_1', 'r_eff_1','n_1','ellip_1', 'r_eff_2','n_2','ellip_2','theta'],
        ['xc','yc','flux'],
        ['xc','yc','flux','r_eff','ellip','theta'],
        ['xc','yc','flux','r_eff','ellip','theta'],]

@pytest.mark.parametrize('mask', [None, jnp.zeros((100,100))])
def test_SourceProperties(mask):
    test_im = jnp.ones((100,100))
    SP = SourceProperties(test_im, mask = mask)
    SP.measure_properties()
    attributes = ['sky_guess','sky_guess_err', 'flux_guess','flux_guess_err', 'r_eff_guess','r_eff_guess_err','xc_guess','yc_guess']
    for var in attributes:
        assert hasattr(SP, var)

@pytest.mark.parametrize('prof, var_names', zip(prof_names,prof_vars) )
def test_prior_gen(prof, var_names):
    props = SourceProperties(jnp.ones((100,100)) )
    prior_class = props.generate_prior(profile_type = prof)
    assert prior_class.check_vars(verbose = True)

def test_sky_sampling():
    x = jnp.arange(100)
    X,Y = jnp.meshgrid(x,x)

    prior_class = NoSkyPrior(sky_guess=0, sky_guess_err=1.)
    with seed(rng_seed=1):
        params_1 = prior_class.sample(X,Y)
    assert params_1 == 0

    prior_class = FlatSkyPrior(sky_guess = 0, sky_guess_err = 1)
    with seed(rng_seed=1):
        params_2 = prior_class.sample(X,Y)
    assert params_2.shape == () # Should be single value

    prior_class = TiltedPlaneSkyPrior(sky_guess = 0, sky_guess_err = 1)
    with seed(rng_seed=1):
        params_3 = prior_class.sample(X,Y)
    assert params_3.shape == (100,100) # Should be 2D array 

def test_PySersicSourcePrior():
    prior = PySersicSourcePrior('pointsource')
    prior.set_gaussian_prior('xc', 10.,1.)
    prior.set_gaussian_prior('yc', 10.,1.)
    prior.set_uniform_prior('flux', 0, 100)
    with seed(rng_seed=1):
        params = prior()
    assert params['xc'] == pytest.approx(8.852981, rel=1e-5)
    assert params['yc'] == pytest.approx(8.907836, rel=1e-5)
    assert params['flux'] == pytest.approx(37.12473, rel=1e-5)

def test_PySersicMultiPrior():
    catalog = {}
    catalog['x'] = [10,15,20]
    catalog['y'] = [20,15,10]
    catalog['flux'] = [100,100,100]
    catalog['r'] = [5,5,5]
    catalog['theta'] = [0,0,0]
    catalog['type'] = ['pointsource','exp','sersic']
    mp = PySersicMultiPrior(catalog)
    with seed(rng_seed=1):
        params = mp()
    
    assert params['flux_0'] == pytest.approx(77.05961, abs = 1e-4)
    assert params['xc_0'] == pytest.approx(8.90784, abs = 1e-4)
    assert params['yc_0'] == pytest.approx(19.67145, abs = 1e-4)
    assert params['flux_1'] == pytest.approx(113.12357, abs = 1e-4)
    assert params['xc_1'] == pytest.approx(13.87507, abs = 1e-4)
    assert params['yc_1'] == pytest.approx(16.05198, abs = 1e-4)
    assert params['r_eff_1'] == pytest.approx(9.10913, abs = 1e-4)
    assert params['ellip_1'] == pytest.approx(0.32804, abs = 1e-4)
    assert params['theta_1'] == pytest.approx(6.21555, abs = 1e-4)
    assert params['flux_2'] == pytest.approx(65.84653, abs = 1e-4)
    assert params['xc_2'] == pytest.approx(21.38217, abs = 1e-4)
    assert params['yc_2'] == pytest.approx(11.08138, abs = 1e-4)
    assert params['r_eff_2'] == pytest.approx(3.35782, abs = 1e-4)
    assert params['ellip_2'] == pytest.approx(0.78032, abs = 1e-4)
    assert params['theta_2'] == pytest.approx(0.61594, abs = 1e-4)
    assert params['n_2'] == pytest.approx(4.79926, abs = 1e-4)
    
def test_sky_estimate():
    rng = np.random.default_rng(1234567)
    test_im = rng.normal(size = (100,100))
    med,std,npix = estimate_sky(test_im, n_pix_sample= 5)
    assert med == pytest.approx(0., abs = 1e-2)
    assert std == pytest.approx(1., abs = 1e-2)
    assert npix == 1900


import copy
from typing import Callable, Optional, Tuple, Union

import arviz as az
import asdf
import re
import corner
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import numpyro
import pandas as pd
import xarray
from jax import random
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

from .priors import PySersicMultiPrior, base_profile_params
from .rendering import BaseRenderer

ArrayLike = Union[np.array, jax.numpy.array]
ListLike = Union[np.array,jax.numpy.array,list]

class PySersicResults():
    def __init__(self,
                data: ArrayLike,
                rms: ArrayLike,
                psf: ArrayLike,
                loss_func:Callable,
                renderer: BaseRenderer,
                mask: Optional[ArrayLike] = None,
                ):
        """Initialize Results Object

        Parameters
        ----------
        data : ArrayLike
            data from fitter
        rms : ArrayLike
            rms from fitter
        psf : ArrayLike
            psf from fitter
        loss_func : Callable
            the loss function chosen
        renderer : BaseRenderer
            the renderer created by the fitter
        mask : Optional[ArrayLike], optional
            mask, if it was used in the fitting, by default None
        """
        self.data = data 
        self.rms = rms 
        self.psf = psf 
        self.mask = mask 
        self.loss_func = loss_func 
        self.renderer = renderer


    def __repr__(self)->str:
        if not hasattr(self,'runtype'):
            self.runtype='unknown'
        out = f'PySersicResults object for pysersic fit of type: {self.runtype}\n'
        return out

    def add_method_used(self,method):
        self.svi_method_used = method 


    def add_prior(self,prior):
        """add the prior object to the result object once created

        Parameters
        ----------
        prior : pysersic.priors.BasePrior
            created prior object
        """
        self.prior = prior 
    
    def injest_data(self, 
                sampler: Optional[numpyro.infer.mcmc.MCMC] =  None, 
                svi_res_dict: Optional[dict] =  None,
                purge_extra: Optional[bool] = True,
                num_sample: Optional[int] = 1_000,
                rkey: Optional[jax.random.PRNGKey] = random.PRNGKey(5)
        ) -> pd.DataFrame:
        """Method to injest data from optimized SVI model or results of sampling. 
        When sampling data is input, sets the class attribute `sampling_results`, while for an SVI run, saves `svi_results.`
        A PysersicResults object can have at most data for one SVI and one sampling run. Both class atttributes are 
        instances of arviz.InferenceData objects.

        Parameters
        ----------
        sampler : Optional[numpyro.infer.mcmc.MCMC], optional
            numpyro sampler containing results
        svi_res_dict : Optional[dict], optional
            Dictionary containing 'guide', 'model' and 'svi_result' specifying a trained SVI model
        purge_extra : Optional[bool], optional
            Whether to purge variables containing 'auto', 'base' or 'unwrapped' often used in reparamaterization, by default True
        num_sample: Optional[int]
            Number of samples to draw from trained SVI posterior, no effect if sampling was used.
        rkey : Optional[jax.random.PRNGKey], optional
            PRNG key to use, by default jax.random.PRNGKey(5)

        Returns
        -------
        pandas.DataFrame
            ArviZ Summary of results

        Raises
        ------
        AssertionError
            Must supply one of sampler or svi_dict
        """

        if sampler is None and (svi_res_dict is None):
            raise AssertionError("Must svi results dictionary or sampled sampler")

        elif sampler is not None:
            self.idata = az.from_numpyro(sampler)
            self.input  = sampler
            self.idata = self._parse_injested_data(self.idata,purge_extra=purge_extra)
            self.runtype = 'sampling'
        else:
            assert 'guide' in svi_res_dict.keys()
            assert 'model' in svi_res_dict.keys()
            assert 'svi_result' in svi_res_dict.keys()
            self.input  = svi_res_dict
            post_raw = svi_res_dict['guide'].sample_posterior(rkey, svi_res_dict['svi_result'].params, sample_shape = ((num_sample,)))
            #Convert to arviz
            post_dict = {}
            for key in post_raw:
                post_dict[key] = post_raw[key][jnp.newaxis,]
            self.idata = az.from_dict(post_dict)
            self.idata = self._parse_injested_data(self.idata,purge_extra=purge_extra)
            self.runtype='svi'

        return

    def _parse_injested_data(self,data:az.InferenceData, purge_extra:bool = True, save_model: bool = True)->az.InferenceData:
        """Helper function to postprocess the poterior object (internal use).

        Parameters
        ----------
        data : arviz.InferenceData
            _description_
        purge_extra : bool, optional
            whether to purge extra params not part of the fitting, by default True
        save_model : bool
            Whether to set self.models with model images from posterior
        Returns
        -------
        arviz.InferenceData
            the cleaned up object
        """
        var_names = list( data.posterior.data_vars )
        
        for var in var_names:
            if 'theta' in var:
                new_theta = np.remainder(data['posterior'][var]+np.pi, np.pi)
                data['posterior'][var] = new_theta

        if purge_extra:
            to_drop = []
            for var in var_names:
                if ('base' in var) or ('auto' in var) or ('unwrapped' in var):
                    to_drop.append(var)
                elif 'model' in var:
                    to_drop.append(var)
                    if save_model:
                        self.models = data['posterior'][var]
            data.posterior = data.posterior.drop_vars(to_drop).drop_dims(['model_dim_0','model_dim_1'], errors = 'ignore')
        return data


    def summary(self)->pd.DataFrame:
        """Convenience function for returning the summary dataframe using the arviz summary.

        Returns
        -------
        pandas.DataFrame
            data frame containing the arviz summary of the fit.
        """
        return az.summary(self.idata)
    

    def get_median_model(self,):
        print('This function has been deprecated as it was buggy and not implemented well. If you require a model we suggest using Fit[x].find_map() of looking at the models from the posterior in PySersicResults.models')
        print('Please reach out if you have any questions.')
        return NotImplementedError

    def corner(self,quantiles=[.16,.50,.84,],**kwargs):
        """Return a corner plot of the parameter estimation

        Parameters
        ----------
        quantiles: ListLike, optional  
            which quantiles to mark on the corner plot, can pass None, by default [0.16,0.5,0.84]
        **kwargs:
            any additional arguments to pass to corner (for the single plot case).
        Returns
        -------
        matplotlib.figure
            fig object containing the corner plots
        """

        return corner.corner(self.idata,show_titles=True,quantiles=quantiles,**kwargs)

        
    def retrieve_param_quantiles(self,
                                quantiles:ListLike=[0.16,0.5,0.84],
                                return_dataframe:bool=False)->Union[pd.DataFrame,dict]:
        """retrieve quantiles on the parameter estimation

        Parameters
        ----------
        quantiles : ListLike, optional
            array of quantiles to pull, must be between 0 and 1, by default [0.16,0.5,0.84]
        return_dataframe : bool, optional
            whether to return dataframe instead of simple dict, by default False

        Returns
        -------
        Union[pd.DataFrame,dict]
            dict or dataframe with index/keys as parameters and columns/values as the chosen quantiles.
        """
        xx = self.idata.posterior.quantile(quantiles, dim = ['chain','draw']).to_dict()
        out = {} 
        for i in xx['data_vars'].keys():
            out[i] = xx['data_vars'][i]['data']
        if return_dataframe==False:
            return out
        else:
            df = pd.DataFrame.from_dict(out).T
            df.columns = quantiles
            return df 

    def retrieve_med_std(self,return_dataframe:bool=False)->Union[pd.DataFrame,dict]:
        out = {}
        q_dict = self.retrieve_param_quantiles()
        for key in q_dict.keys():
            qs = q_dict[key]
            if isinstance(qs[0], list):
                qs = np.array(qs)
            med = qs[1]
            std = 0.5*np.abs(qs[2] - qs[0])
            out[key] = np.array([med,std])
        if return_dataframe:
            out = pd.DataFrame.from_dict(out).T.rename(columns= {0:'median',1:'std'})
        return out


    def latex_table(self,quantiles:ListLike=[0.16,0.5,0.84]):
        """
        Generate a simple AASTex deluxetable with the fit parameters. Prints the result.

        Parameters
        ----------
        quantiles : ListLike, optional
            quantiles to use must be len 3 as we do upper-median and median-lower to get +/- values, by default [0.16,0.5,0.84]

        Raises
        ------
        AssertionError
            if the quantile list does not have three values 
        """
        out = "\\begin{deluxetable}{lr}[b]\n"
        out+= "\\tablehead{\n"
        out+= "\colhead{Parameter} & \colhead{\hspace{4.5cm}Value\hspace{.5cm}}}\n"
        out+="\caption{Best Fit Parameters for Pysersic Fit}\n"
        out+="\startdata \n"
        df = self.retrieve_param_quantiles(quantiles=quantiles,return_dataframe=True)
        if len(df.columns)!=3:
            raise AssertionError('Must Choose 3 quantile positions for +/- calculation')
        for i in df.index:
            q0 = quantiles[0]
            q1 = quantiles[1]
            q2 = quantiles[2]
            plus = df.loc[i,q2] - df.loc[i,q1]
            minus = df.loc[i,q1] - df.loc[i,q0]
            if '_' in i:
                use = i.split('_')[0] + f"_{{\\rm {i.split('_')[1]}}}"
            else:
                use=i
            if i =='theta':
                use = r'\theta'
            out+=f"{use} & {df.loc[i,0.50]:.3f}_{{-{minus:.3f}}}^{{+{plus:.3f}}} \\\\ \n"
        out+="\enddata \n"
        out+="\end{deluxetable}"
        print(out)

    def get_chains(self)->xarray.Dataset:
        """
        Wrapper for az.extract, producing the chains/draws for the run

        Returns
        -------
        xarray.Dataset
            chain object
        """

        return az.extract(self.idata)

    
    def compute_statistic(self,parameter:str,func:Callable,)->ArrayLike:
        """
        Compute an arbitrary array statistic on the chain for a given parameter.
        For example, the std of all ellipticity draws. or the mean of the sersic draws

        Parameters
        ----------
        parameter : str
            a legal parameter from the fit. Use e.g. `results.svi_summary()` to see them.
        func : Callable
            any function which reads in array-like data and computes something

        Returns
        -------
        ArrayLike
            the computed statistic(s)
        """
        chain = self.get_chains()
        return func(chain[parameter]).data
    
    
    def save_result(self,fname:str):
        """Save summary of the fit, copies of the data, the chains, and some other info about priors and renderers
        into an asdf file for later retrieval. 

        Parameters
        ----------
        fname : str
            filename to save to.
        """
        tree = {} 
        tree['input_data'] = {} 
        tree['input_data']['image'] = np.array(self.data)
        tree['input_data']['rms'] = np.array(self.rms)
        tree['input_data']['psf'] = np.array(self.psf)
        tree['input_data']['mask'] = np.array(self.mask)
        tree['loss_func'] = str(self.loss_func)
        tree['rendere_type'] = str( type(self.renderer) )
        tree['method_used'] = self.runtype
        if self.runtype == 'svi':
            tree['svi_method_used'] = self.svi_method_used
        tree['prior_info'] = self.prior.__str__()
        tree['posterior'] = self.idata.to_dict()['posterior']
        for i in tree['posterior']:
            i = np.array(i)
        af = asdf.AsdfFile(tree=tree)
        if not fname.endswith('.asdf'):
            fname+='.asdf'
        af.write_to(fname)

    def sample_posterior(self, num_sample: int, purge_extra: Optional[bool] = True, rkey: Optional[jax.random.PRNGKey] = random.PRNGKey(7))-> az.InferenceData:
        """Generate extra samples from an trained SVI posterior

        Parameters
        ----------
        num_sample : int
            number of samples to draw
        purge_extra : Optional[bool], optional
            Whether to purge variables containing 'auto', 'base' or 'unwrapped' often used in reparamaterization, by default True
        rkey : Optional[jax.random.PRNGKey], optional
            PRNG key to use, by default jax.random.PRNGKey(7)

        Returns
        -------
        az.InferenceData
            arviz InferenceData object containing posterior
        """
        assert self.runtype == 'svi', "Can only add samples if SVI was used for inference"
        post_raw = self.input['guide'].sample_posterior(rkey, self.input['svi_result'].params, sample_shape = ((num_sample,)))
        #Convert to arviz
        post_dict = {}
        for key in post_raw:
            post_dict[key] = post_raw[key][jnp.newaxis,]
        idata = az.from_dict(post_dict)
        idata = self._parse_injested_data(idata,purge_extra=purge_extra, save_model= False)
        return idata

def parse_multi_results(results: PySersicResults, source_num: int) -> PySersicResults:
    """Function written to parse results from a FitMulti instance to isolate a single source. A new PySersicResults class is created with only the posteriors of the specified source. The original chains saved under `.idata_all`

    Parameters
    ----------
    results : PySersicResults
        Results class from a FitMulti instance
    source_num : int 
        Source number to isolate or if equal to -1 will reset to the joint posterior of all sources

    Returns
    -------
    PySersicResults
        Results class with all the meta-data the same but the specified source isolated. The original posterior is saved under `.idata_all`    
    """
    new_res = copy.copy(results)

    assert type(results.prior) == PySersicMultiPrior , "Results must be from a FitMulti instance"
    
    if source_num == -1:
        if not hasattr(new_res, 'idata_all'):
           print("No need to reset posterior, returning original")
        else:
            idata_all = copy.copy(new_res.idata_all)
            new_res.__setattr__('idata', idata_all)
            new_res.__delattr__('idata_all')
    else:
        #Select variables for source
        prof_type = new_res.prior.catalog['type'][source_num]
        param_names = base_profile_params[prof_type]
        source_names = [pname + f'_{source_num}' for pname in param_names]

        #Search for other meta variables like sky etc.
        meta_names = []
        for k in new_res.idata.posterior.keys():
            if re.search(f"_[0-{new_res.prior.N_sources:d}]", k) is None:
                meta_names.append(k)

        if hasattr(new_res, 'idata_all'):
            idata = new_res.idata_all
        else:
            idata = new_res.idata
            new_res.__setattr__('idata_all', idata)

        
        post_source = az.extract(idata, var_names = source_names+meta_names , combined = False)
        idata_source = az.InferenceData(posterior = post_source)
        idata_source.rename_vars(dict(zip(source_names,param_names)), inplace = True)

        new_res.__delattr__('idata')
        new_res.__setattr__('idata', idata_source)
    return new_res


def get_bounds(im: np.array ,scale: float) -> Tuple[float,float]:
    """Generate bounds based on image mean and standard deviation

    Parameters
    ----------
    im : np.array
        images
    scale : float
        Number of +/- sigmas for bounds

    Returns
    -------
    Tuple[float,float]
       Bounds to use
    """
    m = np.mean(im)
    s = np.std(im)
    vmin = m - scale*s 
    vmax = m+scale*s 
    return vmin, vmax

def plot_image(image: np.array,mask: np.array,sig: np.array,psf: np.array,cmap:str ='gray_r',scale:float = 2.0,size:float = 8.) -> Tuple[plt.Figure,plt.Axes]:
    """Plot a summary figure with the image, sigma map and psf side by side

    Parameters
    ----------
    image : np.array
        Image to plot
    mask : np.array
        Mask
    sig : np.array
        sigma or noise map
    psf : np.array
        Point spread function
    cmap : str, optional
        Color map to use, by default 'gray_r'
    scale : float, optional
        Number of +/- std's of image to make the bounds, by default 2.0
    size : float, optional
        Size of figure, will be size*3 x size, by default 8.

    Returns
    -------
    Tuple[plt.Figure,plt.Axes]
        Figure and axes objects
    """
    im_ratio = image.shape[0]/image.shape[1]
    fig, ax = plt.subplots(1,3,figsize=(size*3,size*im_ratio))
    masked_image = np.ma.masked_array(image,mask)
    masked_sigma = np.ma.masked_array(sig,mask)
    im_vmin, im_vmax = get_bounds(masked_image,scale)
    sig_vmin,sig_vmax = get_bounds(masked_sigma,scale)
    psf_vmin,psf_vmax = get_bounds(psf,scale)
    ax[0].imshow(masked_image,origin='lower',cmap=cmap,vmin=im_vmin,vmax=im_vmax)
    ax[1].imshow(masked_sigma,origin='lower',cmap=cmap,vmin=sig_vmin,vmax=sig_vmax)
    ax[2].imshow(psf,origin='lower',cmap=cmap,vmin=psf_vmin,vmax=psf_vmax)
    return fig, ax


def plot_residual(image: np.array,model: np.array,mask: np.array = None,scale:float =2.0,cmap:str ='gray_r',colorbar:bool =True,**resid_plot_kwargs)-> Tuple[plt.Figure,plt.Axes]:
    """Generate a summary plot comparing the data to the best fit model

    Parameters
    ----------
    image : np.array
        Original image
    model : np.array
        Best fit model image
    mask : np.array, optional
        Pixel by Pixel mask, by default None
    scale : float, optional
        Number of +/- std's to make the bounds of the image, by default 2.0
    cmap : str, optional
        color map to use, by default 'gray_r'
    colorbar : bool, optional
        Whether or not to show a color bar, by default True

    Returns
    -------
    Tuple[plt.Figure,plt.Axes]
        Figure and axes objects
    """
    fig, ax = plt.subplots(1,3,figsize=(13,3))
    if mask is not None:
        masked_image = np.ma.masked_array(image,mask)
        masked_model = np.ma.masked_array(model,mask)
    else:
        masked_image = image 
        masked_model = model 
    im_vmin, im_vmax = get_bounds(masked_image,scale)
    ax[0].imshow(masked_image,origin='lower',cmap=cmap,vmin=im_vmin,vmax=im_vmax)
    ax[1].imshow(masked_model,origin='lower',cmap=cmap,vmin=im_vmin,vmax=im_vmax)
    residual = masked_image - masked_model 
    ri = ax[2].imshow(residual,origin='lower',cmap='seismic',**resid_plot_kwargs)
    ax1_divider = make_axes_locatable(ax[2])
    cax1 = ax1_divider.append_axes("right", size="7%", pad="2%")
    cb1 = fig.colorbar(ri, cax=cax1)
    return fig, ax 


from .priors import (PySersicMultiPrior, PySersicSourcePrior, SourceProperties,autoprior)
from .pysersic import FitMulti, FitSingle, check_input_data,parse_mask
from .rendering import FourierRenderer, HybridRenderer, PixelRenderer
from .results import PySersicResults, parse_multi_results



import copy
import warnings
from abc import ABC, abstractmethod
from functools import partial
from typing import Callable, Optional

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
import pandas
import tqdm
from jax.random import PRNGKey
from jax.typing import ArrayLike
from numpyro import deterministic, infer, optim
from numpyro.handlers import condition, trace
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.svi import SVIRunResult

from .exceptions import KernelError, ShapeMatchError
from .loss import gaussian_loss
from .priors import PySersicMultiPrior, PySersicSourcePrior, base_profile_params
from .rendering import BaseRenderer, HybridRenderer, base_profile_types
from .results import PySersicResults


def identity(x: Callable) -> Callable:
    return x


class BaseFitter(ABC):
    """
    Base class for Pysersic Fitters
    """

    def __init__(
        self,
        data: ArrayLike,
        rms: ArrayLike,
        psf: ArrayLike,
        mask: Optional[ArrayLike] = None,
        loss_func: Optional[Callable] = gaussian_loss,
        renderer: Optional[BaseRenderer] = HybridRenderer,
        renderer_kwargs: Optional[dict] = {},
    ) -> None:
        """Initialze BaseFitter class

        Parameters
        ----------
        data : ArrayLike
            Science image to be fit
        weight_map : ArrayLike
            Weight map (one over the variance) corresponding to `data`, must be the same shape
        psf_map : ArrayLike
            Pixelized PSF
        mask : Optional[ArrayLike], optional
            Array specifying the mask, `True` or 1 signifies a pixel should be masked, must be same shape as `data`
        sky_model : Optional[str], optional
            One of None, 'flat' or 'tilted-plane' specifying how to model the sky background
        renderer : Optional[BaseRenderer], optional
            The renderer to be used to generate model images, by default HybridRenderer
        renderer_kwargs : Optional[dict], optional
            Any additional arguments to pass to the renderer, by default {}
        """

        self.loss_func = loss_func

        self.data = jnp.array(data.astype(np.float32))
        self.rms = jnp.array(rms.astype(np.float32))
        self.psf = jnp.array(psf.astype(np.float32))
        self.mask = parse_mask(mask, self.data)
        data_isgood = check_input_data(
            self.data, rms=self.rms, psf=self.psf, mask=jnp.logical_not(self.mask)
        )
        self.renderer = renderer(data.shape, self.psf, **renderer_kwargs)
        self.prior = None
        self.prior_dict = {}

    def set_loss_func(self, loss_func: Callable) -> None:
        """Set loss function to be used for inference

        Parameters
        ----------
        loss_func : Callable
            Functions which takes samples the loss function, see utils/loss.py for some examples.
        """
        self.loss_func = loss_func

    def set_prior(
        self, parameter: str, distribution: numpyro.distributions.Distribution
    ) -> None:
        """Set the prior for a specific parameter

        Parameters
        ----------
        parameter : str
            Parameter to be set
        distribution : numpyro.distributions.Distribution
            Numpyro distribution object corresponding to the prior
        """
        self.prior_dict[parameter] = distribution

    def sample(
        self,
        rkey: jax.random.PRNGKey,
        num_samples: int = 1000,
        num_warmup: int = 1000,
        num_chains: int = 2,
        init_strategy: Optional[Callable] = infer.init_to_sample,
        sampler_kwargs: Optional[dict] = {},
        mcmc_kwargs: Optional[dict] = {},
        return_model: Optional[bool] = True,
        reparam_func: Optional[Callable] = identity,
    ) -> pandas.DataFrame:
        """Perform inference using a NUTS sampler

        Parameters
        ----------
        num_samples : int, optional
            Number of samples to draw, by default 1000
        num_warmup : int, optional
            Number of warmup samples, by default 1000
        num_chains : int, optional
            Number of chains to run, by default 2
        init_strategy : Optional[Callable], optional
            Initialization strategy for the sampler, by default infer.init_to_sample. See numpyro.infer.initialization for more options
        sampler_kwargs : Optional[dict], optional
            Arguments to pass to the numpyro NUTS kernel
        mcmc_kwargs : Optional[dict], optional
            Arguments to pass to the numpyro MCMC sampler
        return_model : Optional[bool]
            Whether to return the model images but adds a small memory/time overhead, by default True
        rkey : Optional[jax.random.PRNGKey], optional
            PRNG key to use, by default jax.random.PRNGKey(3)

        Returns
        -------
        pandas.DataFrame
            ArviZ summary of posterior
        """

        model = self.build_model(return_model=return_model)
        model = reparam_func(model)
        sampler = infer.MCMC(
            infer.NUTS(model, init_strategy=init_strategy, **sampler_kwargs),
            num_chains=num_chains,
            num_samples=num_samples,
            num_warmup=num_warmup,
            **mcmc_kwargs,
        )
        sampler.run(rkey)
        self.sampling_results = PySersicResults(
            data=self.data,
            rms=self.rms,
            psf=self.psf,
            mask=self.mask,
            loss_func=self.loss_func,
            renderer=self.renderer,
        )
        self.sampling_results.add_prior(self.prior)
        self.sampling_results.injest_data(sampler=sampler)
        return self.sampling_results

    def _train_SVI(
        self,
        autoguide: numpyro.infer.autoguide.AutoContinuous,
        method: str,
        rkey: jax.random.PRNGKey,
        ELBO_loss: Optional[Callable] = infer.Trace_ELBO(5),
        lr_init: Optional[int] = 1e-2,
        num_round: Optional[int] = 3,
        SVI_kwargs: Optional[dict] = {},
        train_kwargs: Optional[dict] = {},
        return_model: Optional[bool] = True,
        num_sample: Optional[int] = 1_000,
    ) -> pandas.DataFrame:
        """
        Internal function to perform inference using stochastic variational inference.

        Parameters
        ----------
        autoguide : numpyro.infer.autoguide.AutoContinuous
            Function to build guide
        method: str
            name of method being used; for saving results
        Elbo_loss : Optional[Callable], optional
            Loss function to use, by default infer.Trace_ELBO(1), see numpyro.infer.elbo for more options
        lr_init : Optional[int], optional
            Initial learning rate, by default 1e-2
        num_round : Optional[int], optional
            Number of rounds for training, lr decreases each round, by default 3
        SVI_kwargs : Optional[dict], optional
            Additional arguments to pass to numpyro.infer.SVI, by default {}
        train_kwargs : Optional[dict], optional
            Additional arguments to pass to utils.train_numpyro_svi_early_stop, by default {}
        return_model : Optional[bool]
            Whether to return the model images but adds a small memory/time overhead, by default True
        num_sample: Optional[int]
            Number of samples to draw from trained SVI posterior
        rkey : Optional[jax.random.PRNGKey], optional
            PRNG key, by default jax.random.PRNGKey(6)

        Returns
        -------
        pandas.DataFrame
            ArviZ summary of posterior
        """
        model_cur = self.build_model(return_model=return_model)
        guide = autoguide(model_cur)

        svi_kernel = SVI(
            model_cur, guide, optim.Adam(0.01), loss=ELBO_loss, **SVI_kwargs
        )
        numpyro_svi_result = train_numpyro_svi_early_stop(
            svi_kernel, rkey=rkey, lr_init=lr_init, num_round=num_round, **train_kwargs
        )

        svi_res_dict = dict(guide=guide, model=model_cur, svi_result=numpyro_svi_result)
        self.svi_results = PySersicResults(
            data=self.data,
            rms=self.rms,
            psf=self.psf,
            mask=self.mask,
            loss_func=self.loss_func,
            renderer=self.renderer,
        )
        self.svi_results.injest_data(
            svi_res_dict=svi_res_dict, purge_extra=True, num_sample=num_sample
        )
        self.svi_results.add_prior(self.prior)
        self.svi_results.add_method_used(method)
        return self.svi_results

    def find_MAP(
        self,
        rkey: jax.random.PRNGKey,
        return_model: Optional[bool] = True,
        purge_extra: Optional[bool] = True,
    ):
        """Find the "best-fit" parameters as the maximum a-posteriori and return a dictionary with values for the parameters.

        Parameters
        ----------
        return_model : Optional[bool]
            Whether to return the model images but adds a small memory/time overhead, by default True
        rkey : Optional[jax.random.PRNGKey], optional
            rng key, by default jax.random.PRNGKey(3)

        Returns
        -------
        dict
            dictionary with fit parameters and their values.
        """
        model_cur = self.build_model(return_model=return_model)
        autoguide_map = infer.autoguide.AutoDelta(
            model_cur, init_loc_fn=infer.init_to_median
        )
        train_kwargs = dict(
            lr_init=0.05,
            num_round=3,
            frac_lr_decrease=0.1,
            patience=250,
            optimizer=optim.Adam,
            max_train=int(2e4),
        )
        svi_kernel = SVI(model_cur, autoguide_map, optim.Adam(0.01), loss=Trace_ELBO())

        res = train_numpyro_svi_early_stop(svi_kernel, rkey=rkey, **train_kwargs)

        use_dict = {}
        for key in res.params.keys():
            pref = key.split("_auto_loc")[0]
            use_dict[pref] = res.params[key]
        trace_out = trace(condition(model_cur, use_dict)).get_trace()
        real_out = {}

        if purge_extra:
            for key in trace_out:
                if "Loss" in key:
                    continue
                elif key == "model":
                    real_out[key] = np.asarray(trace_out[key]["value"])
                elif not (
                    "base" in key
                    or "auto" in key
                    or "unwrapped" in key
                    or "factor" in key
                    or "loss" in key
                ):
                    real_out[key] = np.round(trace_out[key]["value"], 5)
        else:
            for k, v in trace_out.items():
                real_out[k] = v["value"]
        return real_out

    def estimate_posterior(
        self,
        rkey: jax.random.PRNGKey,
        method: str = "laplace",
        return_model: bool = True,
        num_sample: Optional[int] = 1_000,
    ) -> pandas.DataFrame:
        """Estimate the posterior using a method other than MCMC sampling. Generally faster than MCMC, but could be less accurate.
        Current Options are:
        - 'laplace'
            - Uses the Laplace approximation, which finds the MAP and then uses a Gaussian approximation to the posterior. The covariance matrix is calculated using the Hessian of the log posterior at the MAP. Generally the fastest method but can lead to numerical problems, especially for fitting several (~> 5) sources since this involves inverting a large matrix.
        - 'svi-mvn'
            - Use variational inference to fit a multivariate normal distribution to the posterior. In practice should give similar results to 'laplace', as both assume the posterior is a multivariate Gaussian but is trained differently and generally less susceptible to the numerical issues discussed above.
        - 'svi-flow'
            - Uses a neural flow (currently a BNAF, https://arxiv.org/abs/1904.04676) to approximate the posterior. This is more flexible than the two methods above as the flow is a more flexible representation of the posterior which can capture non-Gaussian behavior.However it is slower to train. Also optimization can be inconsistent so use and interpret with caution. Best to cross-reference with sample on tests cases.

        Parameters
        ----------
        method : str, optional
            method to use, by default 'laplace'
        return_model : Optional[bool]
            Whether to return the model images but adds a small memory/time overhead, by default True
                num_sample: Optional[int]
        Number of samples to draw from trained SVI posterior
            Number of samples to draw from trained SVI posterior
        rkey : Optional[jax.random.PRNGKey], optional
            rng key, by default jax.random.PRNGKey(6)

        """
        assert method in ["laplace", "svi-flow", "svi-mvn"]
        if method == "laplace":
            train_kwargs = dict(patience=250, max_train=20000)
            guide_func = partial(
                infer.autoguide.AutoLaplaceApproximation,
                init_loc_fn=infer.init_to_median,
            )
            results = self._train_SVI(
                guide_func,
                method=method,
                lr_init=0.05,
                train_kwargs=train_kwargs,
                return_model=return_model,
                rkey=rkey,
                num_sample=num_sample,
            )
        elif method == "svi-flow":
            train_kwargs = dict(patience=500, max_train=20000)
            guide_func = partial(
                infer.autoguide.AutoBNAFNormal,
                num_flows=1,
                hidden_factors=[16, 8],
                init_loc_fn=infer.init_to_median,
            )
            results = self._train_SVI(
                guide_func,
                method="svi-flow",
                ELBO_loss=infer.Trace_ELBO(8),
                train_kwargs=train_kwargs,
                num_round=3,
                lr_init=5e-2,
                rkey=rkey,
                return_model=return_model,
                num_sample=num_sample,
            )
        elif method == "svi-mvn":
            train_kwargs = dict(patience=200, max_train=5000)
            guide_func = partial(
                infer.autoguide.AutoLowRankMultivariateNormal,
                init_scale=5e-3,
                init_loc_fn=infer.init_to_median,
            )
            results = self._train_SVI(
                guide_func,
                method="svi-mvn",
                ELBO_loss=infer.TraceMeanField_ELBO(5),
                train_kwargs=train_kwargs,
                num_round=3,
                lr_init=5e-2,
                rkey=rkey,
                return_model=return_model,
                num_sample=num_sample,
            )
        return results

    @abstractmethod
    def build_model(self, return_model: bool = True):
        raise NotImplementedError


class FitSingle(BaseFitter):
    """
    Class used to fit a single source
    """

    def __init__(
        self,
        data: ArrayLike,
        rms: ArrayLike,
        psf: ArrayLike,
        prior: PySersicSourcePrior,
        mask: Optional[ArrayLike] = None,
        loss_func: Optional[Callable] = gaussian_loss,
        renderer: Optional[BaseRenderer] = HybridRenderer,
        renderer_kwargs: Optional[dict] = {},
    ) -> None:
        """Initialze FitSingle class

        Parameters
        ----------
        data : ArrayLike
            Science image to be fit
        weight_map : ArrayLike
            Weight map (one over the variance) corresponding to `data`, must be the same shape
        psf_map : ArrayLike
            Pixelized PSF
        mask : Optional[ArrayLike], optional
            Array specifying the mask, `True` or 1 signifies a pixel should be masked, must be same shape as `data`
        sky_model : Optional[str], optional
            One of None, 'flat' or 'tilted-plane' specifying how to model the sky background
        profile_type : Optional[str], optional
            Must be one of: ['sersic','doublesersic','pointsource','exp','dev'] specifying how to paramaterize the source, default 'sersic'
        renderer : Optional[BaseRenderer], optional
            The renderer to be used to generate model images, by default HybridRenderer
        renderer_kwargs : Optional[dict], optional
            Any additional arguments to pass to the renderer, by default {}
        """

        super().__init__(
            data,
            rms,
            psf,
            loss_func=loss_func,
            mask=mask,
            renderer=renderer,
            renderer_kwargs=renderer_kwargs,
        )

        if prior.profile_type not in base_profile_types:
            raise AssertionError("Profile must be one of:", base_profile_types)
        self.profile_type_number = 0
        self.prior = prior

    def build_model(self, return_model: bool = True) -> Callable:
        """Generate Numpyro model for the specified image, profile and priors

        Returns
        -------
        model: Callable
            Function specifying the current model in Numpyro, can be passed to inference algorithms
        """
        # Make sure all variables have priors
        check_prior = self.prior.check_vars
        if not check_prior:
            raise AssertionError(
                "Not all variables have priors, please run .autogenerate_priors"
            )

        @numpyro.handlers.reparam(config=self.prior.reparam_dict)
        def model(return_model: bool = return_model):
            params = self.prior()
            out = self.renderer.render_source(
                params, self.prior.profile_type, suffix=self.prior.suffix
            )

            sky = self.prior.sample_sky(self.renderer.X, self.renderer.Y)

            obs = out + sky
            if return_model:
                deterministic(f"model{self.prior.suffix}", obs)
            self.loss_func(
                obs, self.data, self.rms, self.mask, suffix=self.prior.suffix
            )

        return model


class FitMulti(BaseFitter):
    """
    Class used to fit multiple sources within a single image
    """

    def __init__(
        self,
        data: ArrayLike,
        rms: ArrayLike,
        psf: ArrayLike,
        prior: PySersicMultiPrior,
        mask: Optional[ArrayLike] = None,
        loss_func: Optional[Callable] = gaussian_loss,
        renderer: Optional[BaseRenderer] = HybridRenderer,
        renderer_kwargs: Optional[dict] = {},
    ) -> None:
        """Initialze FitMulti class

        Parameters
        ----------
        data : ArrayLike
            Science image to be fit
        weight_map : ArrayLike
            Weight map (one over the variance) corresponding to `data`, must be the same shape
        psf_map : ArrayLike
            Pixelized PSF
        mask : Optional[ArrayLike], optional
            Array specifying the mask, `True` or 1 signifies a pixel should be masked, must be same shape as `data`
        sky_model : Optional[str], optional
            One of None, 'flat' or 'tilted-plane' specifying how to model the sky background
        profile_type : Optional[str], optional
            Must be one of: ['sersic','doublesersic','pointsource','exp','dev'] specifying how to paramaterize the source, default 'sersic'
        renderer : Optional[BaseRenderer], optional
            The renderer to be used to generate model images, by default HybridRenderer
        renderer_kwargs : Optional[dict], optional
            Any additional arguments to pass to the renderer, by default {}
        """
        super().__init__(
            data,
            rms,
            psf,
            mask=mask,
            loss_func=loss_func,
            renderer=renderer,
            renderer_kwargs=renderer_kwargs,
        )
        self.prior = prior

    def build_model(self, return_model: bool = True) -> Callable:
        """Generate Numpyro model for the specified image, profile and priors

        Returns
        -------
        model: Callable
            Function specifying the current model in Numpyro, can be passed to inference algorithms
        """

        @numpyro.handlers.reparam(config=self.prior.reparam_dict)
        def model(return_model: bool = return_model):
            source_variables = self.prior()

            out = self.renderer.render_for_model(
                source_variables, self.prior.catalog["type"], suffix=self.prior.suffix
            )
            sky = self.prior.sample_sky(self.renderer.X, self.renderer.Y)

            obs = out + sky

            if return_model:
                deterministic(f"model{self.prior.suffix}", obs)
            self.loss_func(
                obs, self.data, self.rms, self.mask, suffix=self.prior.suffix
            )

        return model

    def find_MAP(
        self,
        return_model: Optional[bool] = True,
        rkey: Optional[jax.random.PRNGKey] = jax.random.PRNGKey(3),
        purge_extra: Optional[bool] = True,
    ):
        """Find the "best-fit" parameters as the maximum a-posteriori and return a dictionary with values for the parameters.

        Parameters
        ----------
        return_model : Optional[bool], optional
            whether to return the model image, adds a small time and memory overhead, by default True
        rkey : Optional[jax.random.PRNGKey], optional
            rng key, by default jax.random.PRNGKey(3)

        Returns
        -------
        dict
            dictionary with fit parameters and their values.
        """
        raw_dict = super().find_MAP(
            return_model=return_model, rkey=rkey, purge_extra=purge_extra
        )

        if purge_extra:
            results_dict = {}
            for i in range(self.prior.N_sources):
                results_dict[f"source_{i}"] = {}
                for pname in base_profile_params[self.prior.catalog["type"][i]]:
                    results_dict[f"source_{i}"][pname] = raw_dict.pop(pname + f"_{i:d}")
            results_dict.update(raw_dict)
        else:
            results_dict = raw_dict
        return results_dict


def train_numpyro_svi_early_stop(
    svi_class: SVI,
    num_round: Optional[int] = 3,
    max_train: Optional[int] = 5000,
    lr_init: Optional[float] = 0.01,
    frac_lr_decrease: Optional[float] = 0.1,
    patience: Optional[int] = 100,
    optimizer: Optional[optim._NumPyroOptim] = optim.Adam,
    rkey: Optional[PRNGKey] = PRNGKey(10),
) -> SVIRunResult:
    """Optimize a SVI model by training for multiple rounds with a deacreasing learning rate, and early stopping for each round

    Parameters
    ----------
    svi_class : SVI
        Initialized numpyo SVI class, note that the optimizer will be overwritten
    num_round : Optional[int], optional
        Number of training rounds, by default 3
    max_train : Optional[int], optional
        Max number of training epochs per ropund, by default 3000
    lr_init : Optional[float], optional
        Initial learning rate, by default 0.1
    frac_lr_decrease : Optional[float], optional
        Multiplicative factor to change learning rate each round, by default 0.1
    patience : Optional[int], optional
        Number of training epochs to wait for improvement, by default 100
    optimizer : Optional[optim._NumPyroOptim], optional
        Optimizer algorithm tro use, by default optim.Adam
    rkey : Optional[PRNGKey], optional
        Jax PRNG key, by default PRNGKey(10)

    Returns
    -------
    SVIRunResult
        SVI Result class containing trained model
    """
    optim_init = optimizer(lr_init)
    svi_class.__setattr__("optim", optim_init)

    init_state = svi_class.init(rkey)
    all_losses = []

    @partial(jax.jit, static_argnums=1)
    def update_func(state, svi_class, lr):
        svi_class.__setattr__("optim", optimizer(lr))
        state, loss = svi_class.stable_update(state)
        return state, loss

    best_state, best_loss = update_func(init_state, svi_class, lr_init)

    for r in range(num_round):
        losses = []
        wait_counter = 0
        svi_state = copy.copy(best_state)

        if r > 0:
            lr_cur = lr_init * frac_lr_decrease**r
            best_loss = jnp.inf
        else:
            lr_cur = lr_init

        with tqdm.trange(1, max_train + 1) as t:
            for j in t:
                svi_state, loss = update_func(svi_state, svi_class, lr_cur)
                if loss < best_loss:
                    best_loss = loss
                    best_state = copy.copy(svi_state)
                    wait_counter = 0
                elif wait_counter >= patience:
                    break
                else:
                    wait_counter += 1
                t.set_postfix_str(
                    f"Round = {r:d},step_size = {lr_cur:.1e} loss: {best_loss:.3e}",
                    refresh=False,
                )
                losses.append(loss)

        all_losses.append(losses)

    return SVIRunResult(svi_class.get_params(best_state), svi_state, losses)


def parse_mask(mask: ArrayLike = None, data: ArrayLike = None):
    if mask is None:
        return jnp.ones_like(data).astype(jnp.bool_)
    else:
        return jnp.logical_not(jnp.array(mask.astype(int))).astype(jnp.bool_)


def check_input_data(
    data: ArrayLike, rms: ArrayLike, psf: ArrayLike, mask: ArrayLike = None
):
    """Check input data for certain conditions and raise warnings or exceptions if needed

    Parameters
    ----------
    data : ArrayLike
        input image (galaxy/cutout)
    rms : ArrayLike
       rms/error map of the data.
    psf : ArrayLike
        pixelized PSF
    mask :ArrayLike, optional
       mask with True/1 indicating a pixel should be masked, by default None

    Raises
    ------
    RMSWarning
        If the rms map is highly discrepant from the rms of the data
    PSFNormalizationWarning
        if the pst normalization is not ~1
    KernelError
        if the provided PSF is larger than the input image (exception)
    MaskWarning
        If more than 70% of the image is masked.
    """
    rms = jnp.array(rms)
    data = jnp.array(data)
    psf = jnp.array(psf)
    if data.shape != rms.shape:
        raise ShapeMatchError("RMS map ndims must match input data ndims.")
    if jnp.median(rms) > 5 * jnp.std(data):
        warnings.warn(
            "Input RMS map appears to be highly offset (>5x) in magnitude from the rms of the pixels in the input image."
        )
    if not jnp.isclose(jnp.sum(psf), 1.0, 0.1):
        warnings.warn(
            "PSF does not appear to be appropriately normalized; Sum(psf) is more than 0.1 away from 1."
        )
    if jnp.all(data.shape < psf.shape):
        raise KernelError("PSF pixel image size must be smaller than science image.")
    if mask is not None:
        mask = parse_mask(mask, data)
        if jnp.sum(mask) / jnp.prod(jnp.array(mask.shape)) < 0.7:
            warnings.warn(
                "More than 70 percent of input image is masked. Is this correct? (Pysersic treats True/1 as masked; you may need to flip your boolean array.) "
            )
        if mask.shape != data.shape:
            raise ShapeMatchError("Mask ndims must match input data ndims.")
    return True
